from typing import Any, Callable, Optional, Union, TYPE_CHECKING

from ._omm_stream import _OMMStream
from ._stream_factory import create_omm_stream
from .base_stream import StreamOpenWithUpdatesMixin
from ..._content_type import ContentType  # noqa
from ..._core.session import get_valid_session
from ..._tools import cached_property, create_repr, make_callback

if TYPE_CHECKING:
    from ... import OpenState
    from ..._core.session import Session
    from .contrib import ContribResponse, ContribType
    from ..._types import ExtendedParams, OptStr, Strings


class OMMStream(StreamOpenWithUpdatesMixin):
    """
    Open an OMM stream.

    Parameters
    ----------
    session: Session
        The Session defines the source where you want to retrieve your data

    name: string
        RIC to retrieve item stream.

    api: string, optional
        specific name of RDP streaming defined in config file.
        i.e. 'streaming.trading-analytics.endpoints.redi'
        Default: 'streaming.trading-analytics.endpoints.redi'

    domain: string
        Specify item stream domain (MarketPrice, MarketByPrice, ...)
        Default : "MarketPrice"

    service: string, optional
        Specify the service to subscribe on.
        Default: None

    fields: string or list, optional
        Specify the fields to retrieve.
        Default: None

    extended_params: dict, optional
        Specify optional params
        Default: None

    Raises
    ------
    Exception
        If request fails or if Refinitiv Services return an error

    Examples
    --------
    >>> import refinitiv.data as rd
    >>> APP_KEY = "app_key"
    >>> session = rd.session.desktop.Definition(app_key=APP_KEY).get_session()
    >>> session.open()
    >>>
    >>> euro = rd.delivery.omm_stream.Definition("EUR=").get_stream(session)
    >>> euro.open()
    >>>
    >>> def on_update_callback(stream, msg):
    ...     print(msg)
    >>>
    >>> definition = rd.delivery.omm_stream.Definition("THB=")
    >>> thb = definition.get_stream(session)
    >>> thb.on_update(on_update_callback)
    >>> thb.open()
    """

    def __init__(
        self,
        name: str,
        session: Optional["Session"] = None,
        api: "OptStr" = None,
        domain: str = "MarketPrice",
        service: "OptStr" = None,
        fields: Optional["Strings"] = None,
        extended_params: "ExtendedParams" = None,
    ) -> None:
        self._session = get_valid_session(session)
        self._always_use_default_session = session is None
        self._name = name
        self._api = api
        self._domain = domain
        self._service = service
        self._fields = fields
        self._extended_params = extended_params

    @cached_property
    def _stream(self) -> _OMMStream:
        return create_omm_stream(
            ContentType.NONE,
            api=self._api,
            session=self._session,
            name=self._name,
            domain=self._domain,
            service=self._service,
            fields=self._fields,
            extended_params=self._extended_params,
        )

    @property
    def status(self):
        status = {
            "status": self._stream.state,
            "code": self._stream.stream_state,
            "message": self._stream.message_state,
        }
        return status

    def open(self, with_updates: bool = True) -> "OpenState":
        """
        Opens the OMMStream to start to stream.
        Once it's opened, it can be used in order to retrieve data.

        Parameters
        ----------
        with_updates : bool, optional
            actions:
                True - the streaming will work as usual
                        and the data will be received continuously.
                False - only one data snapshot will be received
                        (single Refresh 'NonStreaming') and
                        stream will be closed automatically.

            Defaults to True

        Returns
        -------
        OpenState
            current state of this OMM stream object.

        Examples
        --------
        >>> from refinitiv.data.delivery import omm_stream
        >>> definition = omm_stream.Definition("EUR")
        >>> stream = definition.get_stream()
        >>> stream.open()
        """
        return super().open(with_updates=with_updates)

    async def open_async(self, with_updates: bool = True) -> "OpenState":
        """
        Opens asynchronously the OMMStream to start to stream

        Parameters
        ----------
        with_updates : bool, optional
            actions:
                True - the streaming will work as usual
                        and the data will be received continuously.
                False - only one data snapshot will be received
                        (single Refresh 'NonStreaming') and
                        stream will be closed automatically.

        Returns
        -------
        OpenState
            current state of this OMM stream object.

        Examples
        --------
        >>> from refinitiv.data.delivery import omm_stream
        >>> definition = omm_stream.Definition("EUR")
        >>> stream = definition.get_stream()
        >>> await stream.open_async()
        """
        return await super().open_async(with_updates=with_updates)

    def close(self) -> "OpenState":
        """
        Closes the OMMStream connection, releases resources

        Returns
        -------
        OpenState
            current state of this OMM stream object.

        Examples
        --------
        >>> from refinitiv.data.delivery import omm_stream
        >>> definition = omm_stream.Definition("EUR")
        >>> stream = definition.get_stream()
        >>> stream.open()
        >>> stream.close()
        """
        return super().close()

    def on_refresh(self, func: Callable[[dict, "OMMStream"], Any]) -> "OMMStream":
        """
        This function called when the stream is opened or
        when the record is refreshed with a new image.
        This callback receives a full image.

        Parameters
        ----------
        func : Callable, optional
             Callable object to process retrieved refresh data

        Returns
        -------
        OMMStream
            current instance is an OMM stream object.

        Examples
        --------
        Prerequisite: The default session must be opened.
        >>> from datetime import datetime
        >>> from refinitiv.data.delivery import omm_stream
        >>>
        >>> def display_response(event, stream):
        ...      print(f'Refresh received at {datetime.now}')
        ...      print(event)
        >>>
        >>> definition = omm_stream.Definition("EUR")
        >>> stream = definition.get_stream()
        >>> stream.on_refresh(display_response)
        >>>
        >>> stream.open()
        """
        self._stream.on_refresh(make_callback(func))
        return self

    def on_update(self, func: Callable[[dict, "OMMStream"], Any]) -> "OMMStream":
        """
        This function called when an update is received.

        Parameters
        ----------
        func : Callable, optional
            Callable object to process retrieved update data

        Returns
        -------
        OMMStream
            current instance is an OMM stream object.

        Examples
        --------
        Prerequisite: The default session must be opened.
        >>> from datetime import datetime
        >>> from refinitiv.data.delivery import omm_stream
        >>>
        >>> def display_response(event, stream):
        ...     print(f'Update received at {datetime.now}')
        ...     print(event)
        >>>
        >>> definition = omm_stream.Definition("EUR")
        >>> stream = definition.get_stream()
        >>> stream.on_update(display_response)
        >>>
        >>> stream.open()
        """
        self._stream.on_update(make_callback(func))
        return self

    def on_status(self, func: Callable[[dict, "OMMStream"], Any]) -> "OMMStream":
        """
        This function these notifications are emitted when
        the status of one of the requested instruments changes

        Parameters
        ----------
        func : Callable, optional
            Callable object to process retrieved status data

        Returns
        -------
        OMMStream
            current instance is an OMM stream object.

        Examples
        --------
        Prerequisite: The default session must be opened.
        >>> from datetime import datetime
        >>> from refinitiv.data.delivery import omm_stream
        >>>
        >>> def display_response(event, stream):
        ...     print(f'Status received at {datetime.now}')
        ...     print(event)
        >>>
        >>> definition = omm_stream.Definition("EUR")
        >>> stream = definition.get_stream()
        >>> stream.on_status(display_response)
        >>>
        >>> stream.open()
        """
        self._stream.on_status(make_callback(func))
        return self

    def on_complete(self, func: Callable[[dict, "OMMStream"], Any]) -> "OMMStream":
        """
        This function called on complete event

        Parameters
        ----------
        func : Callable, optional
            Callable object to process when retrieved on complete data.

        Returns
        -------
        OMMStream
            current instance is an OMM stream object.

        Examples
        --------
        Prerequisite: The default session must be opened.
        >>> from datetime import datetime
        >>> from refinitiv.data.delivery import omm_stream
        >>>
        >>> def display_response(event, stream):
        ...     print(f'Complete received at {datetime.now}')
        ...     print(event)
        >>>
        >>> definition = omm_stream.Definition("EUR")
        >>> stream = definition.get_stream()
        >>> stream.on_complete(display_response)
        >>>
        >>> stream.open()
        """
        self._stream.on_complete(make_callback(func))
        return self

    def on_error(self, func: Callable[[dict, "OMMStream"], Any]) -> "OMMStream":
        """
        This function called when an error occurs

        Parameters
        ----------
        func : Callable, optional
            Callable object to process when retrieved error data.

        Returns
        -------
        OMMStream
            current instance is an OMM stream object.

        Examples
        --------
        Prerequisite: The default session must be opened.
        >>> from datetime import datetime
        >>> from refinitiv.data.delivery import omm_stream
        >>>
        >>> def display_response(event, response):
        ...     print(f'Error received at {datetime.now}')
        ...     print(event)
        >>>
        >>> definition = omm_stream.Definition("EUR")
        >>> stream = definition.get_stream()
        >>> stream.on_error(display_response)
        >>>
        >>> stream.open()
        """
        self._stream.on_error(make_callback(func))
        return self

    def on_ack(self, on_ack: Callable[[dict, "OMMStream"], Any]) -> "OMMStream":
        """
        This function called when the stream received an ack message after sending a contribution .

        Parameters
        ----------
        on_ack : Callable, optional
             Callable object to process retrieved ack data

        Returns
        -------
        OMMStream
            current instance is an OMM stream object.

        Examples
        --------
        Prerequisite: The default session must be opened.
        >>> from datetime import datetime
        >>> from refinitiv.data.delivery import omm_stream
        >>>
        >>> def display_response(response, event_type, event):
        ...     print(f'{response} - {event_type} received at {datetime.now}')
        ...     print(event)
        >>>
        >>> definition = omm_stream.Definition("EUR")
        >>> stream = definition.get_stream()
        >>> stream.on_ack(lambda event, stream: display_response(stream, 'ack', event))
        >>>
        >>> stream.open()
        >>> stream.contribute({"ASK": 123, "BID": 125})
        """
        self._stream.on_ack(make_callback(on_ack))
        return self

    def __repr__(self):
        return create_repr(
            self,
            middle_path="omm_stream",
            class_name=self.__class__.__name__,
        )

    def contribute(
        self,
        fields: dict,
        contrib_type: Union[str, "ContribType", None] = None,
        post_user_info: Optional[dict] = None,
    ) -> "ContribResponse":
        """
        Function to send OnStream contribution request.

        Parameters
        ----------
        fields: dict{field:value}
            Specify fields and values to contribute.

        contrib_type: Union[str, ContribType], optional
            Define the contribution type
            Default: "Update"

        post_user_info: dict, optional
            PostUserInfo object Represents information about the posting user.
                Address: string, required
                    Dotted-decimal string representing the IP Address of the posting user.
                UserID: int, required
                    Specifies the ID of the posting user

        Returns
        -------
        ContribResponse

        Examples
        --------
        Prerequisite: The default session must be opened.
        >>> from datetime import datetime
        >>> from refinitiv.data.delivery import omm_stream
        >>>
        >>> def display_response(response, event_type, event):
        ...     print(f'{response} - {event_type} received at {datetime.now}')
        ...     print(event)
        >>>
        >>> definition = omm_stream.Definition("EUR")
        >>> stream = definition.get_stream()
        >>> stream.on_ack(lambda event, stream: display_response(stream, 'ack', event))
        >>>
        >>> stream.open()
        >>> response = stream.contribute({"ASK": 123, "BID": 125})
        """
        return self._stream.contribute(
            fields, contrib_type=contrib_type, post_user_info=post_user_info
        )

    async def contribute_async(
        self,
        fields: dict,
        contrib_type: Union[str, "ContribType", None] = None,
        post_user_info: Optional[dict] = None,
    ) -> "ContribResponse":
        """
        Function to send asynchronous OnStream contribution request.

        Parameters
        ----------
        fields: dict{field:value}
            Specify fields and values to contribute.

        contrib_type: Union[str, ContribType], optional
            Define the contribution type
            Default: "Update"

        post_user_info: dict, optional
            PostUserInfo object Represents information about the posting user.
                Address: string, required
                    Dotted-decimal string representing the IP Address of the posting user.
                UserID: int, required
                    Specifies the ID of the posting user

        Returns
        -------
        ContribResponse

        Examples
        --------
        Prerequisite: The default session must be opened.
        >>> from datetime import datetime
        >>> from refinitiv.data.delivery import omm_stream
        >>>
        >>> def display_response(response, event_type, event):
        ...     print(f'{response} - {event_type} received at {datetime.now}')
        ...     print(event)
        >>>
        >>> definition = omm_stream.Definition("EUR")
        >>> stream = definition.get_stream()
        >>> stream.on_ack(lambda event, stream: display_response(stream, 'ack', event))
        >>>
        >>> stream.open()
        >>> response = await stream.contribute_async({"ASK": 123, "BID": 125})
        """
        return await self._stream.contribute_async(
            fields, contrib_type=contrib_type, post_user_info=post_user_info
        )



# coding: utf8

from typing import Optional
from .._object_definition import ObjectDefinition
from ...._tools import create_repr, try_copy_to_list

from ._enums import RiskType, AssetClass
from ._zc_curve_definition import ZcCurveDefinition
from ...._types import OptStr, Strings


class ZcCurveDefinitions(ObjectDefinition):
    """
    Parameters
    ----------
    index_name : str, optional

    index_tenors : string, optional
        Defines expected rate surface tenor/slices Defaults to the tenors available,
        based on provided market data
    main_constituent_asset_class : AssetClass, optional

    pivot_curve_definition : ZcCurveDefinition, optional

    reference_curve_definition : ZcCurveDefinition, optional

    risk_type : RiskType, optional

    currency : str, optional
        The currency code of the interest rate curve
    discounting_tenor : str, optional
        Mono currency discounting tenor
    id : str, optional
        Id of the curve definition
    name : str, optional
        The name of the interest rate curve
    source : str, optional

    """

    def __init__(
        self,
        index_name: OptStr = None,
        index_tenors: Strings = None,
        main_constituent_asset_class: Optional[AssetClass] = None,
        pivot_curve_definition: Optional[ZcCurveDefinition] = None,
        reference_curve_definition: Optional[ZcCurveDefinition] = None,
        risk_type: Optional[RiskType] = None,
        currency: OptStr = None,
        discounting_tenor: OptStr = None,
        id: OptStr = None,
        name: OptStr = None,
        source: OptStr = None,
    ) -> None:
        super().__init__()
        self.index_name = index_name
        self.index_tenors = try_copy_to_list(index_tenors)
        self.main_constituent_asset_class = main_constituent_asset_class
        self.pivot_curve_definition = pivot_curve_definition
        self.reference_curve_definition = reference_curve_definition
        self.risk_type = risk_type
        self.currency = currency
        self.discounting_tenor = discounting_tenor
        self.id = id
        self.name = name
        self.source = source

    def __repr__(self):
        return create_repr(
            self,
            middle_path="curves.zc_curves",
            class_name=self.__class__.__name__,
        )

    @property
    def index_tenors(self):
        """
        Defines expected rate surface tenor/slices Defaults to the tenors available,
        based on provided market data
        :return: list string
        """
        return self._get_list_parameter(str, "indexTenors")

    @index_tenors.setter
    def index_tenors(self, value):
        self._set_list_parameter(str, "indexTenors", value)

    @property
    def main_constituent_asset_class(self):
        """
        :return: enum AssetClass
        """
        return self._get_enum_parameter(AssetClass, "mainConstituentAssetClass")

    @main_constituent_asset_class.setter
    def main_constituent_asset_class(self, value):
        self._set_enum_parameter(AssetClass, "mainConstituentAssetClass", value)

    @property
    def pivot_curve_definition(self):
        """
        :return: object ZcCurveDefinition
        """
        return self._get_object_parameter(ZcCurveDefinition, "pivotCurveDefinition")

    @pivot_curve_definition.setter
    def pivot_curve_definition(self, value):
        self._set_object_parameter(ZcCurveDefinition, "pivotCurveDefinition", value)

    @property
    def reference_curve_definition(self):
        """
        :return: object ZcCurveDefinition
        """
        return self._get_object_parameter(ZcCurveDefinition, "referenceCurveDefinition")

    @reference_curve_definition.setter
    def reference_curve_definition(self, value):
        self._set_object_parameter(ZcCurveDefinition, "referenceCurveDefinition", value)

    @property
    def risk_type(self):
        """
        :return: enum RiskType
        """
        return self._get_enum_parameter(RiskType, "riskType")

    @risk_type.setter
    def risk_type(self, value):
        self._set_enum_parameter(RiskType, "riskType", value)

    @property
    def currency(self):
        """
        The currency code of the interest rate curve
        :return: str
        """
        return self._get_parameter("currency")

    @currency.setter
    def currency(self, value):
        self._set_parameter("currency", value)

    @property
    def discounting_tenor(self):
        """
        Mono currency discounting tenor
        :return: str
        """
        return self._get_parameter("discountingTenor")

    @discounting_tenor.setter
    def discounting_tenor(self, value):
        self._set_parameter("discountingTenor", value)

    @property
    def id(self):
        """
        Id of the curve definition
        :return: str
        """
        return self._get_parameter("id")

    @id.setter
    def id(self, value):
        self._set_parameter("id", value)

    @property
    def index_name(self):
        """
        :return: str
        """
        return self._get_parameter("indexName")

    @index_name.setter
    def index_name(self, value):
        self._set_parameter("indexName", value)

    @property
    def name(self):
        """
        The name of the interest rate curve
        :return: str
        """
        return self._get_parameter("name")

    @name.setter
    def name(self, value):
        self._set_parameter("name", value)

    @property
    def source(self):
        """
        :return: str
        """
        return self._get_parameter("source")

    @source.setter
    def source(self, value):
        self._set_parameter("source", value)



from typing import TYPE_CHECKING, Union, List

if TYPE_CHECKING:
    import httpx


class ParsedData:
    def __init__(
        self,
        status: dict,
        raw_response: "httpx.Response",
        content_data: Union[dict, list, str] = None,
        error_codes: Union[str, int, List[int]] = None,
        error_messages: Union[str, List[str]] = None,
    ) -> None:
        self.status = status
        self.raw_response = raw_response
        self.content_data = content_data
        if isinstance(error_codes, (int, str)):
            error_codes = [error_codes]
        self._error_codes = error_codes or []
        if isinstance(error_messages, str):
            error_messages = [error_messages]
        self._error_messages = error_messages or []

    @property
    def error_codes(self) -> List[int]:
        return self._error_codes

    @error_codes.setter
    def error_codes(self, value: int):
        if not isinstance(value, list):
            value = [value]
        self._error_codes = value

    @property
    def first_error_code(self) -> int:
        if len(self._error_codes) > 0:
            return self._error_codes[0]
        return 0

    @property
    def error_messages(self) -> List[str]:
        return self._error_messages

    @error_messages.setter
    def error_messages(self, value: str):
        if not isinstance(value, list):
            value = [value]
        self._error_messages = value

    @property
    def first_error_message(self) -> str:
        if len(self._error_messages) > 0:
            return self._error_messages[0]
        return ""

    def as_dict(self) -> dict:
        return {
            "status": self.status,
            "raw_response": self.raw_response,
            "content_data": self.content_data,
            "error_codes": self.error_codes,
            "error_messages": self.error_messages,
        }

    def __eq__(self, o: object) -> bool:
        if isinstance(o, dict):
            return self.as_dict() == o
        return super().__eq__(o)

    def __repr__(self) -> str:
        return str(self.as_dict())



# coding: utf8

from enum import Enum, unique


@unique
class UnderlyingType(Enum):
    """
    Underlying type of the option.
    Possible values:
        - Eti
        - Fx
    """

    ETI = "Eti"
    FX = "Fx"



from enum import Enum, unique


@unique
class InterpolationMode(Enum):
    AKIMA_METHOD = "AkimaMethod"
    CUBIC_DISCOUNT = "CubicDiscount"
    CUBIC_RATE = "CubicRate"
    CUBIC_SPLINE = "CubicSpline"
    FORWARD_MONOTONE_CONVEX = "ForwardMonotoneConvex"
    FRITSCH_BUTLAND_METHOD = "FritschButlandMethod"
    HERMITE = "Hermite"
    KRUGER_METHOD = "KrugerMethod"
    LINEAR = "Linear"
    LOG = "Log"
    MONOTONIC_CUBIC_NATURAL_SPLINE = "MonotonicCubicNaturalSpline"
    MONOTONIC_HERMITE_CUBIC = "MonotonicHermiteCubic"
    TENSION_SPLINE = "TensionSpline"



from typing import TYPE_CHECKING

from ._serializer import Serializer

if TYPE_CHECKING:
    from .._types import OptFloat


class BidAsk(Serializer):
    def __init__(
        self,
        ask: "OptFloat" = None,
        bid: "OptFloat" = None,
    ) -> None:
        super().__init__()
        self.ask = ask
        self.bid = bid

    @property
    def ask(self):
        return self._get_param("ask")

    @ask.setter
    def ask(self, value):
        self._set_param("ask", value)

    @property
    def bid(self):
        return self._get_param("bid")

    @bid.setter
    def bid(self, value):
        self._set_param("bid", value)



import atexit
import json
import os
import re
from json.decoder import WHITESPACE  # noqa
from typing import Any, Optional, Union, List
import sys as _sys

from ._external_libraries import python_configuration as ext_config_mod  # noqa
from pyee import EventEmitter
from watchdog.events import PatternMatchingEventHandler
from watchdog.observers import Observer

from ._errors import RDError
from ._tools._common import get_from_path
from . import _config_defaults

_observer: Optional[Observer] = None


def _enable_watch():
    global _observer
    _observer = Observer()
    _event_handler = _RDPConfigChangedHandler(patterns=_config_files_paths)
    _dir_names = {os.path.dirname(f) for f in _config_files_paths}
    [_observer.schedule(_event_handler, dirname) for dirname in _dir_names]
    _observer.start()

    atexit.register(_disable_watch)


def _disable_watch():
    if _observer:
        _observer.stop()
        _observer.join()


def _dispose():
    _disable_watch()
    get_config().remove_all_listeners()


def _load_config_from_file(filepath):
    _configs = _create_configs([filepath] + _config_files_paths)
    _new_config = ext_config_mod.ConfigurationSet(*_configs)
    get_config().configs = _new_config.configs


def _get_filepath(rootdir, filename):
    if rootdir and filename:
        path = os.path.join(rootdir, filename)
        return path


_SUBS_PATTERN = re.compile(r".*?\${(\w+(-\w+)*(_\w+)*(:\w+)*)}.*?")


def _process_match(value, match, root: dict):
    match = match or []
    for g in match:
        path = g[0]
        old = f"${{{path}}}"
        new = get_from_path(root, path, ":")
        new = None if isinstance(new, list) else new
        value = value.replace(old, new or old)
    return value


def _substitute_values(data: dict, root: dict = None):
    if not data:
        return data

    for k, v in data.items():
        if hasattr(v, "get"):
            _substitute_values(v, root)
        elif isinstance(v, str):
            match = _SUBS_PATTERN.findall(v)
            v = _process_match(v, match, root)
            data[k] = v
    return data


def _read_config_file(path: str) -> dict:
    try:
        with open(path, "r") as f:
            data = json.load(f, cls=_JSONDecoder)
    except FileNotFoundError:
        return {}
    except Exception as exc:  # noqa
        raise RDError(0, f"Error happened during config file {path} read") from exc

    return _substitute_values(data, data)


def _create_configs(files_paths: List[str]):
    config_from_dict = ext_config_mod.config_from_dict
    dicts = [_read_config_file(f) for f in files_paths] + [_config_defaults.config]
    configs = [config_from_dict(d) for d in dicts]
    return configs


def _create_rdpconfig(files_paths) -> "_RDPConfig":
    if isinstance(files_paths, str):
        files_paths = [files_paths]
    configs = _create_configs(files_paths)
    return _RDPConfig(*configs)


class ConfigEvent:
    UPDATE = "update"


class _RDPConfigChangedHandler(PatternMatchingEventHandler):
    def on_any_event(self, event):
        configs = _create_configs(_config_files_paths)
        _new_config = ext_config_mod.ConfigurationSet(*configs)
        config = get_config()
        if _new_config.as_dict() != config.as_dict():
            config.configs = _new_config.configs
            try:
                config.emit(ConfigEvent.UPDATE)
            except Exception:  # noqa
                pass


class _JSONDecoder(json.JSONDecoder):
    _ENV_SUBS_PATTERN = re.compile(r".*?\${(\w+)}.*?")

    def decode(self, s, _w=WHITESPACE.match):
        match = self._ENV_SUBS_PATTERN.findall(s)
        if match:
            for g in match:
                s = s.replace(f"${{{g}}}", os.environ.get(g, g))
            s = s.replace("\\", "\\\\")
        return super().decode(s, _w)


class _keys:  # noqa
    endpoints = "endpoints"
    log_level = "logs.level"
    log_filename = "logs.transports.file.name"
    log_file_enabled = "logs.transports.file.enabled"
    log_console_enabled = "logs.transports.console.enabled"
    log_file_size = "logs.transports.file.size"
    log_filter = "logs.filter"
    log_max_files = "logs.transports.file.maxFiles"
    log_interval = "logs.transports.file.interval"
    watch_enabled = "config-change-notifications-enabled"

    http_request_timeout = "http.request-timeout"
    http_auto_retry_config = "http.auto-retry"
    http_max_connections = "http.max-connections"
    http_max_keepalive_connections = "http.max-keepalive-connections"

    desktop_sessions = "sessions.desktop"

    @staticmethod
    def desktop_session(session_name) -> str:
        return "sessions.desktop.%s" % session_name

    @staticmethod
    def desktop_base_uri(session_name) -> str:
        return "sessions.desktop.%s.base-url" % session_name

    @staticmethod
    def desktop_platform_paths(session_name) -> str:
        return "sessions.desktop.%s.platform-paths" % session_name

    @staticmethod
    def desktop_handshake_url(session_name) -> str:
        return "sessions.desktop.%s.handshake-url" % session_name

    @staticmethod
    def desktop_endpoints(session_name) -> str:
        return "sessions.desktop.%s.endpoints" % session_name

    platform_sessions = "sessions.platform"
    platform_session_default_server_mode = "sessions.platform.default.server-mode"

    @staticmethod
    def platform_session(session_name) -> str:
        return "sessions.platform.%s" % session_name

    @staticmethod
    def platform_endpoints(session_name) -> str:
        return "sessions.platform.%s.endpoints" % session_name

    @staticmethod
    def platform_base_uri(session_name) -> str:
        return "sessions.platform.%s.base-url" % session_name

    @staticmethod
    def platform_auth_uri(session_name) -> str:
        return "sessions.platform.%s.auth.url" % session_name

    @staticmethod
    def platform_token_uri(session_name) -> str:
        return "sessions.platform.%s.auth.token" % session_name

    @staticmethod
    def platform_auto_reconnect(session_name) -> str:
        return "sessions.platform.%s.auto-reconnect" % session_name

    @staticmethod
    def platform_server_mode(session_name) -> str:
        return "sessions.platform.%s.server-mode" % session_name

    @staticmethod
    def platform_realtime_distribution_system(session_name) -> str:
        return "sessions.platform.%s.realtime-distribution-system" % session_name

    @staticmethod
    def stream_connects_locations(streaming_name, endpoint_name) -> str:
        return f"apis.streaming.{streaming_name}.endpoints.{endpoint_name}.locations"

    @staticmethod
    def stream_protocols(streaming_name, endpoint_name) -> str:
        return f"apis.streaming.{streaming_name}.endpoints.{endpoint_name}.protocols"

    @staticmethod
    def get_stream_direct_url(streaming_name, endpoint_name) -> str:
        return f"apis.streaming.{streaming_name}.endpoints.{endpoint_name}.direct-url"


keys = _keys


class _RDPConfig(ext_config_mod.ConfigurationSet):
    def __init__(self, *configs):
        super().__init__(*configs)
        self._emitter = EventEmitter()
        setattr(self, "on", self._emitter.on)
        setattr(self, "remove_listener", self._emitter.remove_listener)
        setattr(self, "remove_all_listeners", self._emitter.remove_all_listeners)
        setattr(self, "emit", self._emitter.emit)
        setattr(self, "listeners", self._emitter.listeners)

    def count(self, event: str = None):
        """
        Returns count of EventEmitter listeners

        Parameters
        ----------
            event: str
                Event name.

        Returns
        -------
        Count of listeners for event or all events listeners
        """
        if event:
            return len(self._emitter.listeners(event))

        count = 0
        for event in self._emitter.event_names():
            count += len(self._emitter.listeners(event))
        return count

    def set_param(self, param: str, value: Any, auto_create: bool = False) -> None:
        """
        Set param to the config.

        Parameters
        ----------
            param: str
                Parameter name.

            value: Any
                Parameter value.

            auto_create: bool
                Default value: False. We use auto_create to create new field in config

        Raises
        ----------
        Exception
            Raise exception if param name is not type 'str' .
        """
        if not isinstance(param, str):
            raise TypeError("Invalid type of parameter name, should be string")

        if not auto_create and not self._has_param(param):
            raise ValueError(
                f"'{param}' is not defined in config. "
                f"To create new parameter please use 'auto_create' property."
            )

        self[param] = value

    def _has_param(self, param: str):
        return param in self

    def get_param(self, param: str) -> Any:
        """
        Get param to the config.

        Parameters
        ----------
            param: str
                Parameter name.

        Raises
        ----------
        Exception
            Raise exception if config doesn't has param.
        """

        try:
            return self[param]
        except KeyError:
            raise AttributeError(f"Config object doesn't has '{param}' attribute")

    def _set_config_index(
        self, index: int, config_: ext_config_mod.Configuration
    ) -> None:
        """
        Set config by using index.
        This method is using for overriding default config.

        Parameters
        ----------
            index: int
                Parameter name.
            config_: Configuration
                Config object
        """
        if config_ is None:
            self._configs.pop(index)

        else:
            self._configs.insert(index, config_)

    def copy(self) -> "_RDPConfig":
        configs = [config.copy() for config in self._configs]
        return _RDPConfig(*configs)


load = _load_config_from_file


def get(key: str, default: Any = None) -> Union[dict, Any]:
    return get_config().get(key, default)


def get_bool(item: str) -> bool:
    return get_config().get_bool(item)


def get_str(item: str, fmt: str = "{}") -> str:
    return get_config().get_str(item, fmt)


def get_int(item: str) -> int:
    return get_config().get_int(item)


def get_list(item: str) -> List[Any]:
    return get_config().get_list(item)


def get_param(param: str) -> Any:
    return get_config().get_param(param)


def get_config() -> "_RDPConfig":
    if _config is None:
        reload()
    return _config


def set_param(param: str, value: Any, auto_create: bool = False) -> None:
    get_config().set_param(param, value, auto_create)


class _defaults:  # noqa
    http_request_timeout: Optional[str] = None
    log_level: Optional[int] = None
    platform_server_mode: Optional[bool] = None


defaults: _defaults = _defaults()

_RD_LIB_CONFIG_PATH = "RD_LIB_CONFIG_PATH"
_RDPLIB_ENV_DIR = "RDPLIB_ENV_DIR"
_default_config_file_name = "refinitiv-data.config.json"

_config: Optional["_RDPConfig"] = None
_config_files_paths: Optional[List[str]] = None
_project_config_dir: Optional[str] = None


def reload():
    global _config, _config_files_paths, _observer, defaults, _project_config_dir

    _custom_filepath: str = os.environ.get(_RD_LIB_CONFIG_PATH, "")

    if _custom_filepath:
        _custom_filepath = os.path.join(_custom_filepath, _default_config_file_name)

    _project_config_dir = os.environ.get(_RDPLIB_ENV_DIR) or os.getcwd()
    _config_files_paths = [
        c
        for c in [
            # CONFIG PROVIDED BY USER
            _custom_filepath,
            # PROJECT_CONFIG_FILE
            _get_filepath(
                rootdir=_project_config_dir, filename=_default_config_file_name
            ),
            # USER_CONFIG_FILE
            _get_filepath(
                rootdir=os.path.expanduser("~"), filename=_default_config_file_name
            ),
        ]
        if c
    ]
    _config = _create_rdpconfig(_config_files_paths)
    _config.load = _load_config_from_file

    _observer = None
    _watch_enabled = _config.get_bool(keys.watch_enabled)
    if _watch_enabled:
        _enable_watch()

    _default_config = _config.configs[len(_config.configs) - 1]

    defaults = _defaults()
    defaults.http_request_timeout = _default_config.get(keys.http_request_timeout)
    from ._log import convert_log_level, root_logger
    import refinitiv.data as rd

    defaults.log_level = convert_log_level(_default_config.get(keys.log_level))
    defaults.platform_server_mode = _default_config.get(
        keys.platform_session_default_server_mode
    )
    log_debug(root_logger(), rd.__version__, _config_files_paths)


def log_debug(logger, version: str, path_list: list):
    logger.debug(f"RD version is {version}; Python version is {_sys.version}")

    try:
        import pkg_resources as _pkg_resources

        _installed_packages = _pkg_resources.working_set
        _installed_packages = sorted(
            [f"{i.key}=={i.version}" for i in _installed_packages]
        )
        logger.debug(
            f"Installed packages ({len(_installed_packages)}): "
            f"{','.join(_installed_packages)}"
        )
    except Exception as e:
        logger.debug(f"Cannot log installed packages, {e}")

    logger.debug(f'Read configs: {", ".join(path_list)}')



from refinitiv.data.content.ipa._curves._bond_curves import RequestItem
from refinitiv.data.content.ipa._curves._bond_curves._curves import (
    CreditCurveParameters,
    CreditCurveDefinition,
)
from refinitiv.data.content.ipa._curves._bond_curves._enums import (
    InterestCalculationMethod,
    BasisSplineSmoothModel,
    CalendarAdjustment,
    CalibrationModel,
    BusinessSector,
    CurveSubType,
    EconomicSector,
    Industry,
    IndustryGroup,
    IssuerType,
    MainConstituentAssetClass,
    Rating,
    RatingScaleSource,
    ReferenceEntityType,
    Seniority,
    InterpolationMode,
    PriceSide,
)
from refinitiv.data.content.ipa._curves._enums import CompoundingType
from refinitiv.data.content.ipa._enums._extrapolation_mode import ExtrapolationMode


def test_credit_curve_request_item():
    # given
    expected_curve_parameters = {
        "interestCalculationMethod": "Dcb_30E_360_ISMA",
        "basisSplineSmoothModel": "AndersonSmoothingSplineModel",
        "calendarAdjustment": "Calendar",
        "calendars": ["string1", "string2"],
        "calibrationModel": "BasisSpline",
        "compoundingType": "Compounded",
        "extrapolationMode": "Constant",
        "interpolationMode": "AkimaMethod",
        "priceSide": "Ask",
        "basisSplineKnots": 10,
        "returnCalibratedParameters": True,
        "useDelayedDataIfDenied": True,
        "useDurationWeightedMinimization": True,
        "useMultiDimensionalSolver": True,
        "valuationDate": "string",
    }

    expected_curve_definition = {
        "businessSector": "AcademicAndEducationalServices",
        "creditCurveTypeFallbackLogic": ["string1", "string2"],
        "curveSubType": "BondCarry",
        "curveTenors": ["string1", "string2"],
        "economicSector": "AcademicAndEducationalServices",
        "industry": "AdvancedMedicalEquipmentAndTechnology",
        "industryGroup": "AerospaceAndDefense",
        "issuerType": "Agency",
        "mainConstituentAssetClass": "Bond",
        "rating": "A",
        "ratingScaleSource": "DBRS",
        "referenceEntityType": "BondIsin",
        "seniority": "SeniorNonPreferred",
        "country": "string",
        "currency": "string",
        "id": "string",
        "name": "string",
        "referenceEntity": "string",
        "source": "string",
    }

    # when
    curve_definition = CreditCurveDefinition(
        business_sector=BusinessSector.ACADEMIC_AND_EDUCATIONAL_SERVICES,
        credit_curve_type_fallback_logic=["string1", "string2"],
        curve_sub_type=CurveSubType.BOND_CARRY,
        curve_tenors=["string1", "string2"],
        economic_sector=EconomicSector.ACADEMIC_AND_EDUCATIONAL_SERVICES,
        industry=Industry.ADVANCED_MEDICAL_EQUIPMENT_AND_TECHNOLOGY,
        industry_group=IndustryGroup.AEROSPACE_AND_DEFENSE,
        issuer_type=IssuerType.AGENCY,
        main_constituent_asset_class=MainConstituentAssetClass.BOND,
        rating=Rating.A,
        rating_scale_source=RatingScaleSource.DBRS,
        reference_entity_type=ReferenceEntityType.BOND_ISIN,
        seniority=Seniority.SENIOR_NON_PREFERRED,
        country="string",
        currency="string",
        id="string",
        name="string",
        reference_entity="string",
        source="string",
    )

    curve_parameters = CreditCurveParameters(
        interest_calculation_method=InterestCalculationMethod.DCB_30_E_360_ISMA,
        basis_spline_smooth_model=BasisSplineSmoothModel.ANDERSON_SMOOTHING_SPLINE_MODEL,
        calendar_adjustment=CalendarAdjustment.CALENDAR,
        calendars=["string1", "string2"],
        calibration_model=CalibrationModel.BASIS_SPLINE,
        compounding_type=CompoundingType.COMPOUNDED,
        extrapolation_mode=ExtrapolationMode.CONSTANT,
        interpolation_mode=InterpolationMode.AKIMA_METHOD,
        price_side=PriceSide.ASK,
        basis_spline_knots=10,
        return_calibrated_parameters=True,
        use_delayed_data_if_denied=True,
        use_duration_weighted_minimization=True,
        use_multi_dimensional_solver=True,
        valuation_date="string",
    )

    # when
    testing_obj = RequestItem(
        curve_definition=curve_definition,
        curve_parameters=curve_parameters,
        curve_tag="curve_tag",
    )

    # then
    assert testing_obj.get_dict() == {
        "curveDefinition": expected_curve_definition,
        "curveParameters": expected_curve_parameters,
        "curveTag": "curve_tag",
    }

    assert testing_obj.curve_definition == curve_definition
    assert testing_obj.curve_parameters == curve_parameters
    assert testing_obj.curve_tag == "curve_tag"



import allure
import pytest

import refinitiv.data.eikon as ek
from refinitiv.data.errors import RDError
from tests.integration.eikon.get_data.conftest import (
    check_dataframe_contains_instruments,
    check_dataframe_contains_fields,
    check_success_response,
)


@allure.suite("Eikon Legacy module - get_data")
@allure.feature("Eikon Legacy module - get_data")
@allure.severity(allure.severity_level.NORMAL)
class TestGetData:
    @allure.title("Get snapshot with real-time pricing")
    @pytest.mark.caseid("38427361")
    @pytest.mark.parametrize("instruments,fields", [(["EUR=", "JPY="], ["BID", "ASK"])])
    @pytest.mark.smoke
    def test_get_pricing_snapshot(self, setup_app_key, instruments, fields):
        df, err = ek.get_data(instruments=instruments, fields=fields)
        check_success_response(df, err)
        check_dataframe_contains_instruments(df, instruments)
        check_dataframe_contains_fields(df, fields)

    @allure.title("Get fundamental and reference data")
    @pytest.mark.caseid("38427362")
    @pytest.mark.parametrize(
        "instruments,fields,expected_fields",
        [
            (
                ["GOOG.O", "MSFT.O", "FB.O"],
                ["TR.Revenue", "TR.GrossProfit"],
                ["Instrument", "Revenue", "Gross Profit"],
            )
        ],
    )
    def test_get_fundamental_and_reference(
        self, setup_app_key, instruments, fields, expected_fields
    ):
        df, err = ek.get_data(instruments=instruments, fields=fields)
        check_success_response(df, err)
        check_dataframe_contains_instruments(df, instruments)
        check_dataframe_contains_fields(df, expected_fields)

    @allure.title("Get fundamental and reference data with global parameters")
    @pytest.mark.caseid("38427363")
    @pytest.mark.parametrize(
        "instruments,fields,parameters,expected_fields",
        [
            (
                ["GOOG.O", "AAPL.O"],
                ["CF_BID", "CF_ASK", "TR.EV", "TR.EVToSales"],
                {"SDate": "0CY", "Curn": "CAD"},
                [
                    "Instrument",
                    "CF_BID",
                    "CF_ASK",
                    "Enterprise Value (Daily Time Series)",
                    "Enterprise Value To Sales (Daily Time Series Ratio)",
                ],
            )
        ],
    )
    def test_get_fundamental_and_reference_with_global_parameters(
        self, setup_app_key, instruments, fields, parameters, expected_fields
    ):
        df, err = ek.get_data(
            instruments=instruments,
            fields=fields,
            parameters=parameters,
        )
        check_success_response(df, err)
        check_dataframe_contains_instruments(df, instruments)
        check_dataframe_contains_fields(df, expected_fields)

    @allure.title("Get fundamental and reference data with data item parameters")
    @pytest.mark.caseid("38427364")
    @pytest.mark.parametrize(
        "instruments,fields,expected_fields",
        [
            (
                ["GOOG.O", "AAPL.O"],
                ["TR.PriceTargetMean(Source=ThomsonReuters)"],
                ["Instrument", "Price Target - Mean"],
            )
        ],
    )
    def test_get_fundamental_and_reference_with_data_item_parameters(
        self, setup_app_key, instruments, fields, expected_fields
    ):
        df, err = ek.get_data(instruments=instruments, fields=fields)
        check_success_response(df, err)
        check_dataframe_contains_instruments(df, instruments)
        check_dataframe_contains_fields(df, expected_fields)

    @allure.title("Get fundamental and reference data with invalid parameters")
    @pytest.mark.caseid("38427365")
    def test_get_data_with_invalid_parameters(self, setup_app_key):
        with pytest.raises(
            RDError,
            match="Error code 1422 | Can not process metadata for request: DataGrid_StandardAsync.requests.instruments - non-empty is required",
        ):
            ek.get_data(instruments=[""], fields=["invalid", ""])



import allure
import pytest

from refinitiv.data.content import custom_instruments as ci
from tests.integration.content.custom_instruments.conftest import (
    instrument_data,
    new_instrument_data,
    get_holidays,
)

basket_dict = {
    "constituents": [
        {"ric": "IBM.N", "weight": 20},
        {"ric": "EPAM.N", "weight": 80},
    ],
    "normalizeByWeight": True,
}

basket_obj = ci.manage.Basket(
    constituents=[
        ci.manage.Constituent(ric="IBM.N", weight=20),
        ci.manage.Constituent(ric="EPAM.N", weight=80),
    ],
    normalize_by_weight=True,
)

new_basket_obj = ci.manage.Basket(
    constituents=[
        ci.manage.Constituent(ric="LSEG.L", weight=50),
        ci.manage.Constituent(ric="VOD.L", weight=30),
        ci.manage.Constituent(ric="MSFT.O", weight=20),
    ],
    normalize_by_weight=False,
)


@allure.suite("Content object - Custom Instrument")
@allure.feature("Basket Custom Instrument")
@allure.severity(allure.severity_level.CRITICAL)
class TestBasketCustomInstrument:
    @allure.title("Create basket Custom Instrument object")
    @pytest.mark.caseid("")
    @pytest.mark.parametrize(
        "basket",
        [
            (basket_dict),
            (basket_obj),
        ],
    )
    def test_create_basket_custom_instrument(
        self, open_session_with_rdp_creds_for_ci, symbol, basket
    ):
        dates_and_calendars_holiday = get_holidays(
            session=open_session_with_rdp_creds_for_ci
        )
        expected_holidays = [
            {"date": "2022-09-01", "reason": "Knowledge Day"},
            {"date": "2022-12-18", "reason": "Hanukkah"},
            {
                "date": dates_and_calendars_holiday.date,
                "reason": dates_and_calendars_holiday.name,
            },
        ]

        cust_inst = ci.manage.create_basket(
            **instrument_data,
            symbol=symbol,
            basket=basket,
            holidays=[
                ci.manage.Holiday(date="2022-09-01", name="Knowledge Day"),
                {"date": "2022-12-18", "reason": "Hanukkah"},
                dates_and_calendars_holiday,
            ],
        )

        for param, value in instrument_data.items():
            assert (
                cust_inst.__getattribute__(param) == value
            ), f"Attribute {param} was not created for {symbol}"
        assert (
            symbol in cust_inst.symbol
        ), f"Created symbol is not the same as requested"
        assert cust_inst.holidays == expected_holidays
        assert cust_inst.basket == basket_obj

        ci.manage.delete(symbol)

    @allure.title("Update Basket Custom Instrument object with params")
    @pytest.mark.caseid("")
    def test_update_basket_custom_instrument(
        self, open_session_with_rdp_creds_for_ci, symbol
    ):
        cust_inst = ci.manage.create_basket(
            **instrument_data, symbol=symbol, basket=basket_dict
        )

        for param, value in new_instrument_data.items():
            cust_inst.__setattr__(param, value)
        cust_inst.basket = new_basket_obj
        cust_inst.save()

        updated_instrument = ci.manage.get(universe=symbol)

        for param, value in new_instrument_data.items():
            assert (
                updated_instrument.__getattribute__(param) == value
            ), f"Attribute {param} was not updated  for {symbol}"
        assert (
            cust_inst.basket == new_basket_obj
        ), f"Changes to Basket obj is not updated"

        ci.manage.delete(symbol)



import allure
import pytest

import refinitiv.data as rd
from refinitiv.data.delivery import cfs
from tests.integration.delivery.cfs.file_downloader.conftest import (
    check_file_in_folder,
    check_file_in_working_folder,
)


@allure.suite("Delivery object - CFS File downloader")
@allure.feature("Delivery object - CFS File downloader")
@allure.severity(allure.severity_level.CRITICAL)
class TestCfsFileDownloader:
    @allure.title("Download and unzip file by specific path")
    @pytest.mark.caseid("36636830")
    @pytest.mark.parametrize(
        "download_path,unzip_path",
        [
            (r"linux_path_type/download_here", r"linux_path_type/unzip_here"),
            (r"windows_path_type\download_here", r"windows_path_type\unzip_here"),
        ],
    )
    def test_download_and_unzip_file_by_specific_path(
        self,
        get_smallest_file_from_bucket,
        open_platform_session,
        download_path,
        unzip_path,
    ):
        file = get_smallest_file_from_bucket

        file_downloader = cfs.file_downloader.Definition(file).retrieve()
        file_downloader.download(path=download_path)
        file_downloader.extract(path=unzip_path)

        check_file_in_folder(file["filename"].replace(":", ""), download_path)
        check_file_in_folder(file["filename"][:-3].replace(":", ""), unzip_path)

    @allure.title("Download file with closed session")
    @pytest.mark.caseid("36636833")
    def test_download_file_with_closed_session(self, get_smallest_file_from_bucket):
        file = get_smallest_file_from_bucket
        session = rd.session.desktop.Definition(app_key=" ").get_session()
        rd.session.set_default(session)

        with pytest.raises(
            ValueError, match="Session is not opened. Can't send any request"
        ):
            cfs.file_downloader.Definition(file).retrieve()

    @allure.title("Download and unzip file without path")
    @pytest.mark.caseid("36636832")
    def test_download_and_unzip_file_without_path(
        self, get_smallest_file_from_bucket, open_platform_session
    ):
        file = get_smallest_file_from_bucket

        file_downloader = cfs.file_downloader.Definition(file).retrieve()
        file_downloader.download()
        file_downloader.extract()

        check_file_in_working_folder(file["filename"].replace(":", ""))



import os

from tests.integration.constants_list import HttpStatusCode


def check_file_downloaded_response(response, expected_file_name):
    actual_file_name = response.data.files[0].file_location
    status_code = response.data.files[0].raw["status_code"]
    is_success = response.data.files[0].is_success

    assert is_success
    assert status_code == HttpStatusCode.TWO_HUNDRED
    assert (
        expected_file_name in actual_file_name
    ), f"actual file name is {actual_file_name}"


def check_file_is_downloaded(filename):
    for root, dirs, files in os.walk("./"):
        for file in files:
            if file.startswith(filename):
                assert True, f"File {filename} not found"
                os.remove(file)
                break


def get_callback(callback_response):
    status_code = callback_response.data.files[0].raw["status_code"]
    assert status_code == HttpStatusCode.TWO_HUNDRED



import pytest

from refinitiv.data.content.filings import search


def test_search_definition_attributes():
    # given
    # when
    s = search.Definition()

    # then
    assert hasattr(s, "query")
    assert hasattr(s, "variables")



from refinitiv.data.content import pricing
from tests.unit.conftest import StubSession
from .conftest import streaming_price_record_json
from ...fincoders_layer.conftest import assert_datetime_dtype_for_df


def test_incorrect_work_pricing_stream_get_snapshot_method():
    # given
    session = StubSession(is_open=True)
    definition = pricing.Definition(["GBP=", "EUR="])
    stream = definition.get_stream(session)
    streaming_price = stream._stream._stream_by_name["GBP="]
    streaming_price._record = streaming_price_record_json

    # when
    df = stream.get_snapshot("GBP=", ["BID"])

    # then
    assert not df.empty


def test_pricing_stream_get_snapshot_columns_datetime64():
    # given
    session = StubSession(is_open=True)
    definition = pricing.Definition(["GBP=", "EUR="])
    stream = definition.get_stream(session)
    streaming_price = stream._stream._stream_by_name["GBP="]
    streaming_price._record = streaming_price_record_json

    # when
    df = stream.get_snapshot("GBP=")

    # then
    assert_datetime_dtype_for_df(df)
    assert sum(i == "datetime64[ns]" for i in df.dtypes) == 32, df.select_dtypes(
        include=["datetime64"]
    ).columns


def test_stream_class_name():
    # given
    expected_class_name = "PricingStream"

    session = StubSession(is_open=True)
    definition = pricing.Definition(["GBP=", "EUR="])
    stream = definition.get_stream(session)

    # when
    class_name = str(type(stream["EUR="]))
    class_name = class_name.rsplit(".", 1)[-1][:-2]

    # then
    assert class_name == expected_class_name



from typing import TYPE_CHECKING

from ...._object_definition import ObjectDefinition


if TYPE_CHECKING:
    from ......_types import OptFloat


class FxForwardTurnFields(ObjectDefinition):
    """
    Creates the Cross Currency curve definition with the definition provided.

    Parameters
    ----------
    bid : float, optional

    ask : float, optional

    """

    def __init__(self, bid: "OptFloat" = None, ask: "OptFloat" = None) -> None:
        super().__init__()
        self.bid = bid
        self.ask = ask

    @property
    def ask(self):
        """
        :return: float
        """
        return self._get_parameter("ask")

    @ask.setter
    def ask(self, value):
        self._set_parameter("ask", value)

    @property
    def bid(self):
        """
        :return: float
        """
        return self._get_parameter("bid")

    @bid.setter
    def bid(self, value):
        self._set_parameter("bid", value)



import numpy as np


def check_response_dataframe_instrument_contains_expected_universes(
    response, expected_universes
):
    dataframe = response.data.df
    try:
        retrieved_universes = dataframe["instrument"].tolist()
    except KeyError:
        retrieved_universes = dataframe["Instrument"].tolist()
    retrieved_universes = set(retrieved_universes)

    if all([isinstance(item, np.int64) for item in retrieved_universes]):
        retrieved_universes = list(map(str, retrieved_universes))

    if isinstance(expected_universes, str):
        assert (
            expected_universes in retrieved_universes
        ), f"Expected universe {expected_universes} is not found in list of instruments retrieved from dataframe: {retrieved_universes}"
    elif isinstance(expected_universes, list):
        for universe in expected_universes:
            assert (
                universe in retrieved_universes
            ), f"Expected universe {universe} is not found in list of instruments retrieved from dataframe: {retrieved_universes}"
    else:
        print(f"Unexpected universe type provided: {type(expected_universes)}")


def get_expected_column(use_field_names_in_headers, expected_titles, expected_names):
    expected_column = expected_titles
    if use_field_names_in_headers:
        expected_column = expected_names

    return expected_column



from functools import partial
from typing import Union

from ._enums import (
    adjustments_arg_parser,
    market_sessions_arg_parser,
    event_types_arg_parser,
)
from ..._content_type import ContentType
from .._historical_data_provider import get_fields_summaries, get_fields_events
from .._intervals import (
    DayIntervalType,
    Intervals,
    get_day_interval_type,
    interval_arg_parser,
)
from ..._tools import (
    urljoin,
    ParamItem,
    ValueParamItem,
    is_date_true,
)
from ..._tools._datetime import hp_datetime_adapter
from ...delivery._data._data_provider import RequestFactory

content_type_by_day_interval_type = {
    DayIntervalType.INTER: ContentType.HISTORICAL_PRICING_INTERDAY_SUMMARIES,
    DayIntervalType.INTRA: ContentType.HISTORICAL_PRICING_INTRADAY_SUMMARIES,
}


def get_content_type_by_interval(
    interval: Union[str, Intervals, DayIntervalType]
) -> ContentType:
    day_interval_type = get_day_interval_type(interval)
    return content_type_by_day_interval_type.get(day_interval_type)


def check_count(value):
    if value is not None and value < 1:
        raise ValueError("Count minimum value is 1")
    return value


hp_summaries_query_params = [
    ValueParamItem(
        "interval",
        function=interval_arg_parser.get_str,
    ),
    ValueParamItem("start", function=hp_datetime_adapter.get_str, is_true=is_date_true),
    ValueParamItem("end", function=hp_datetime_adapter.get_str, is_true=is_date_true),
    ValueParamItem(
        "adjustments",
        function=partial(adjustments_arg_parser.get_str, delim=","),
    ),
    ValueParamItem(
        "sessions",
        function=partial(market_sessions_arg_parser.get_str, delim=","),
    ),
    ValueParamItem("count", function=check_count),
    ParamItem("fields", function=get_fields_summaries),
]

hp_events_query_params = [
    ValueParamItem("interval", function=interval_arg_parser.get_str),
    ValueParamItem(
        "event_types", "eventTypes", partial(event_types_arg_parser.get_str, delim=",")
    ),
    ValueParamItem("start", function=hp_datetime_adapter.get_str, is_true=is_date_true),
    ValueParamItem("end", function=hp_datetime_adapter.get_str, is_true=is_date_true),
    ValueParamItem(
        "adjustments", function=partial(adjustments_arg_parser.get_str, delim=",")
    ),
    ValueParamItem("count", function=check_count),
    ParamItem("fields", function=get_fields_events),
]


class HistoricalPricingRequestFactory(RequestFactory):
    def get_url(self, *args, **kwargs):
        url = args[1]
        url = urljoin(url, "/{universe}")
        return url

    def get_path_parameters(self, session=None, *, universe=None, **kwargs):
        if universe is None:
            return {}
        return {"universe": universe}

    def extend_body_parameters(self, body_parameters, extended_params=None, **kwargs):
        return None

    def extend_query_parameters(self, query_parameters, extended_params=None):
        if extended_params:
            query_parameters = dict(query_parameters)
            query_parameters.update(extended_params)
            for key in ("start", "end"):
                if key in extended_params:
                    arg_date = query_parameters[key]
                    query_parameters[key] = hp_datetime_adapter.get_str(arg_date)
            query_parameters = list(query_parameters.items())

        return query_parameters


class HistoricalPricingEventsRequestFactory(HistoricalPricingRequestFactory):
    @property
    def query_params_config(self):
        return hp_events_query_params


class HistoricalPricingSummariesRequestFactory(HistoricalPricingRequestFactory):
    @property
    def query_params_config(self):
        return hp_summaries_query_params



from itertools import zip_longest
from typing import TYPE_CHECKING, Any

import numpy
import pandas as pd
from pandas import DataFrame

from ..._content_data import Data as ContentData
from ..._content_data_provider import ContentDataProvider
from ..._content_response_factory import ContentResponseFactory
from ..._error_parser import ErrorParser
from ...._tools import ArgsParser, fields_arg_parser, merge_dict_to_dict
from ...._tools._dataframe import convert_df_columns_to_datetime_by_idx, convert_dtypes
from ....delivery._data._data_provider import ContentValidator
from ....delivery._data._data_provider import (
    RequestFactory,
    ValidatorContainer,
)
from ....delivery.endpoint_request import RequestMethod

if TYPE_CHECKING:
    from ....delivery._data._data_provider import ParsedData


# --------------------------------------------------------------------------------------
#   Content data validator
# --------------------------------------------------------------------------------------


class ContractsContentValidator(ContentValidator):
    @classmethod
    def any_valid_content_data(cls, data: "ParsedData") -> bool:
        content_data = data.content_data
        headers = content_data.get("headers", [])
        datas = content_data.get("data", [])
        err_codes = []
        err_msgs = []

        for header, *data_items in zip(headers, *datas):
            header_name = header["name"]

            if "ErrorCode" == header_name:
                err_codes = data_items

            elif "ErrorMessage" == header_name:
                err_msgs = data_items

        counter = len(datas) or 1  # because datas can be empty list
        if err_codes or err_msgs:
            for err_code, err_msg in zip_longest(err_codes, err_msgs, fillvalue=None):
                if err_code or err_msg:
                    counter -= 1
                    data.error_codes.append(err_code)
                    data.error_messages.append(err_msg)

        if counter == 0:
            return False

        return True

    def __init__(self) -> None:
        super().__init__()
        self.validators.append(self.any_valid_content_data)


# ---------------------------------------------------------------------------
#   Content data
# ---------------------------------------------------------------------------


def convert_data_items_to_datetime(df: pd.DataFrame, headers: dict) -> pd.DataFrame:
    columns_indexes = [
        index
        for index, header in enumerate(headers)
        if header.get("type", "") in ["DateTime", "Date"]
    ]

    df = convert_df_columns_to_datetime_by_idx(
        df, columns_indexes, utc=True, delete_tz=True
    )
    return df


def financial_contracts_build_df(raw: dict, **kwargs) -> pd.DataFrame:
    """
    Convert "data" from raw response bond to dataframe format
    """
    data = raw.get("data")
    headers = raw.get("headers")
    if data:
        numpy_array = numpy.array(data, dtype=object)
        df = DataFrame(numpy_array)
        df = convert_data_items_to_datetime(df, headers)
        df.columns = [header["name"] for header in headers]
        df = convert_dtypes(df)
    else:
        df = DataFrame()
    return df


class Data(ContentData):
    """
    This class is designed for storing and managing the response instrument data
    """

    _analytics_headers = None
    _analytics_data = None
    _analytics_market_data = None
    _analytics_statuses = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.raw:
            #   get headers
            self._analytics_headers = self.raw.get("headers")
            #   get data
            self._analytics_data = self.raw.get("data")
            #   get marketData
            self._analytics_market_data = self.raw.get("marketData")
            #   get statuses
            self._analytics_statuses = self.raw.get("statuses")

    @property
    def analytics_headers(self):
        return self._analytics_headers

    @property
    def analytics_data(self):
        return self._analytics_data

    @property
    def analytics_market_data(self):
        return self._analytics_market_data

    @property
    def analytics_statuses(self):
        return self._analytics_statuses

    @property
    def marketdata_df(self):
        """
        Convert "marketData" from raw response bond to dataframe format
        """
        return None


# ---------------------------------------------------------------------------
#   Request factory
# ---------------------------------------------------------------------------


class ContractsRequestFactory(RequestFactory):
    def extend_body_parameters(self, body_parameters, extended_params=None, **kwargs):
        if not extended_params:
            return body_parameters

        if kwargs.get("__plural__") is True:
            body_parameters.update(extended_params)
            return body_parameters

        universes = body_parameters.get("universe", [{}])
        universes[0] = merge_dict_to_dict(universes[0], extended_params)
        return body_parameters

    def get_request_method(self, *, method=None, **kwargs):
        return method or RequestMethod.POST

    def get_body_parameters(
        self,
        *args,
        universe=None,
        definition=None,
        fields=None,
        outputs=None,
        pricing_parameters=None,
        **kwargs,
    ):
        plural = kwargs.get("__plural__")
        if plural is True:
            input_universe = universe
        else:
            input_universe = [definition]

        universe = []
        for item in input_universe:
            item_defn = item
            item_pricing_parameters = not plural and pricing_parameters
            item_extended_params = None

            if hasattr(item, "_kwargs"):
                kwargs = getattr(item, "_kwargs")
                item_defn = kwargs.get("definition")
                item_pricing_parameters = kwargs.get("pricing_parameters")
                item_extended_params = kwargs.get("extended_params")

            inst_defn_dict = item_defn.get_dict()

            if item_extended_params:
                inst_defn_dict.update(item_extended_params)

            data = {
                "instrumentType": item_defn.get_instrument_type(),
                "instrumentDefinition": inst_defn_dict,
            }

            if item_pricing_parameters:
                data["pricingParameters"] = item_pricing_parameters.get_dict()

            universe.append(data)

        body_parameters = {"universe": universe}

        if fields:
            fields = fields_arg_parser.get_list(fields)
            body_parameters["fields"] = fields

        if pricing_parameters and plural is True:
            body_parameters["pricingParameters"] = pricing_parameters.get_dict()

        return body_parameters


def get_data(definition, pricing_parameters=None):
    fields = None
    extended_params = None

    if hasattr(definition, "_kwargs"):
        kwargs = getattr(definition, "_kwargs")
        definition = kwargs.get("definition")
        fields = kwargs.get("fields")
        pricing_parameters = kwargs.get("pricing_parameters")
        extended_params = kwargs.get("extended_params")

    definition_dict = definition.get_dict()

    if extended_params:
        definition_dict.update(extended_params)

    data = {
        "instrumentType": definition.get_instrument_type(),
        "instrumentDefinition": definition_dict,
    }

    if fields:
        fields = fields_arg_parser.get_list(fields)
        data["fields"] = fields

    if pricing_parameters:
        data["pricingParameters"] = pricing_parameters.get_dict()

    return data


def process_bond_instrument_code(code: Any) -> str:
    if code is None or isinstance(code, str):
        return code
    else:
        raise ValueError(
            f"Invalid type of instrument_code, string is expected."
            f"type: {type(code)} is given"
        )


bond_instrument_code_arg_parser = ArgsParser(process_bond_instrument_code)

# ---------------------------------------------------------------------------
#   Data provider
# ---------------------------------------------------------------------------

contracts_data_provider = ContentDataProvider(
    request=ContractsRequestFactory(),
    response=ContentResponseFactory(data_class=Data),
    validator=ValidatorContainer(content_validator=ContractsContentValidator()),
    parser=ErrorParser(),
)



from typing import Union, TYPE_CHECKING

from ._content_response_factory import ContentResponseFactory
from ..delivery._data._data_provider import DataProvider

if TYPE_CHECKING:
    from ..delivery._data._response_factory import ABCResponseFactory
    from ..delivery._data._request_factory import RequestFactory
    from ..delivery._data._validators import ValidatorContainer, BaseValidator
    from ..delivery._data._connection import HttpSessionConnection
    from ..delivery._data._raw_data_parser import Parser


class ContentDataProvider(DataProvider):
    def __init__(
        self,
        connection: "HttpSessionConnection" = None,
        request: "RequestFactory" = None,
        response: "ABCResponseFactory" = None,
        parser: "Parser" = None,
        validator: Union["BaseValidator", "ValidatorContainer"] = None,
    ):
        response = response or ContentResponseFactory()
        super().__init__(
            connection=connection,
            request=request,
            response=response,
            parser=parser,
            validator=validator,
        )



# coding: utf8
from ..._object_definition import ObjectDefinition


class SwaptionMarketDataRule(ObjectDefinition):
    def __init__(self, discount=None, forward=None):
        super().__init__()
        self.discount = discount
        self.forward = forward

    @property
    def discount(self):
        """
        :return: str
        """
        return self._get_parameter("discount")

    @discount.setter
    def discount(self, value):
        self._set_parameter("discount", value)

    @property
    def forward(self):
        """
        :return: str
        """
        return self._get_parameter("forward")

    @forward.setter
    def forward(self, value):
        self._set_parameter("forward", value)



import pytest
from refinitiv.data.content.historical_pricing import events

from refinitiv.data.content import custom_instruments
from tests.unit.conftest import StubSession, args, StubResponse
from . import data_for_tests as td


@pytest.mark.parametrize(
    "input_kwargs, mock_response",
    [
        args(
            input_kwargs={
                "universe": [
                    "S)Batman_9ccfafd6.GESG1-111923",
                ],
            },
            response=td.ONE_UNIVERSE,
        ),
        args(
            input_kwargs={
                "universe": [
                    "S)Batman_9ccfafd6.GESG1-111923",
                ],
                "count": 13000,
            },
            response=td.ONE_UNIVERSE_WITH_TWO_REQUESTS,
        ),
        args(
            input_kwargs={
                "universe": [
                    "S)Batman_9ccfafd6.GESG1-111923",
                    "S)invalid_cust_universe",
                ],
            },
            response=td.TWO_UNIVERSES_ONE_FAILED,
        ),
        args(
            input_kwargs={
                "universe": [
                    "S)Batman_9ccfafd6.GESG1-111923",
                    "S)Batman_3b9a8a02.GESG1-111923",
                ],
            },
            response=td.TWO_UNIVERSES,
        ),
    ],
)
def test_nested_attributes_type_for_response(input_kwargs, mock_response):
    # given
    session = StubSession(is_open=True, response=mock_response)
    definition = custom_instruments.events.Definition(**input_kwargs)

    # when
    testing_response = definition.get_data(session=session)

    # then
    assert isinstance(testing_response.http_status, list)
    assert isinstance(testing_response.http_headers, list)
    assert isinstance(testing_response.request_message, list)
    assert isinstance(testing_response.http_response, list)

    assert all(isinstance(i, list) for i in testing_response.http_status)
    assert all(isinstance(i, list) for i in testing_response.http_headers)
    assert all(isinstance(i, list) for i in testing_response.request_message)
    assert all(isinstance(i, list) for i in testing_response.http_response)


@pytest.mark.asyncio
async def test_get_data_async_df_return_empty_dataframe_if_universe_is_not_found():
    # given
    mock_response = StubResponse(
        [
            {
                "universe": {"ric": "INVALID"},
                "status": {
                    "code": "TS.Intraday.UserRequestError.90001",
                    "message": "The universe is not found.",
                },
            }
        ]
    )
    session = StubSession(is_open=True, response=mock_response)

    # when
    definition = events.Definition("INVALID")
    invalid_response = await definition.get_data_async(session=session)

    # then
    assert invalid_response.data.df.empty is True



from typing import TYPE_CHECKING

from ...._curves._cross_currency_curves._triangulate_definitions import RequestItem
from ......_content_type import ContentType
from ......_tools import create_repr
from ....._content_provider_layer import ContentProviderLayer


if TYPE_CHECKING:
    from ......_types import OptStr, ExtendedParams


class Definition(ContentProviderLayer):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    base_currency : str, optional
        The base currency pair. It is expressed in ISO 4217 alphabetical format
        (e.g., 'EUR').
    base_index_name : str, optional
        The name of the floating rate index (e.g., 'ESTR') applied to the base currency.
    curve_tag : str, optional
        A user-defined string to identify the interest rate curve. it can be used to
        link output results to the curve definition. limited to 40 characters. only
        alphabetic, numeric and '- _.#=@' characters are supported.
    quoted_currency : str, optional
        The quoted currency pair. It is expressed in ISO 4217 alphabetical
        format (e.g., 'USD').
    quoted_index_name : str, optional
        The name of the floating rate index (e.g., 'SOFR') applied to the
        quoted currency.
    valuation_date : str, optional
        The valuation date.
    extended_params : dict, optional
        If necessary other parameters.

    Examples
    --------
    >>> from refinitiv.data.content.ipa.curves._cross_currency_curves import triangulate_definitions
    ... definition = triangulate_definitions.search.Definition(
    ...     base_currency="EUR",
    ...     quoted_currency="CHF"
    ... )
    >>> response = definition.get_data()

    Using get_data_async
    >>> import asyncio
    >>> task = definition.get_data_async()
    >>> response = asyncio.run(task)
    """

    def __init__(
        self,
        base_currency: "OptStr" = None,
        base_index_name: "OptStr" = None,
        curve_tag: "OptStr" = None,
        quoted_currency: "OptStr" = None,
        quoted_index_name: "OptStr" = None,
        valuation_date: "OptStr" = None,
        extended_params: "ExtendedParams" = None,
    ) -> None:

        request_item = RequestItem(
            base_currency=base_currency,
            base_index_name=base_index_name,
            curve_tag=curve_tag,
            quoted_currency=quoted_currency,
            quoted_index_name=quoted_index_name,
            valuation_date=valuation_date,
        )
        super().__init__(
            content_type=ContentType.CROSS_CURRENCY_CURVES_TRIANGULATE_DEFINITIONS_SEARCH,
            universe=request_item,
            extended_params=extended_params,
        )

    def __repr__(self):
        return create_repr(
            self, middle_path="_cross_currency_curves.triangulate_definitions.search"
        )



"""ConfigurationSet class."""

from typing import (
    Any,
    Dict,
    ItemsView,
    Iterable,
    KeysView,
    List,
    Mapping,
    Optional,
    Union,
    ValuesView,
    cast,
)

from .configuration import Configuration
from .helpers import InterpolateEnumType, InterpolateType, clean, interpolate_object


class ConfigurationSet(Configuration):
    """
    Configuration Sets.

    A class that combines multiple :class:`Configuration` instances
    in a hierarchical manner.
    """

    def __init__(
        self,
        *configs: Configuration,
        interpolate: InterpolateType = False,
        interpolate_type: InterpolateEnumType = InterpolateEnumType.STANDARD,
    ):  # noqa: D107
        self._interpolate = {} if interpolate is True else interpolate
        self._interpolate_type = interpolate_type
        try:
            self._configs: List[Configuration] = list(configs)
        except Exception:  # pragma: no cover
            raise ValueError(
                "configs should be a non-empty iterable of Configuration objects"
            )
        if not self._configs:  # pragma: no cover
            raise ValueError(
                "configs should be a non-empty iterable of Configuration objects"
            )
        if not all(
            isinstance(x, Configuration) for x in self._configs
        ):  # pragma: no cover
            raise ValueError(
                "configs should be a non-empty iterable of Configuration objects"
            )
        self._writable = False
        self._default_levels = 1

    def _from_configs(self, attr: str, *args: Any, **kwargs: dict) -> Any:
        last_err = Exception()
        values = []
        for config_ in self._configs:
            try:
                values.append(getattr(config_, attr)(*args, **kwargs))
            except Exception as err:
                last_err = err
                continue
        if not values:
            # raise the last error
            raise last_err
        if all(isinstance(v, Configuration) for v in values):
            result: dict = {}
            for v in values[::-1]:
                result.update(v.as_dict())
            return Configuration(result)
        elif isinstance(values[0], Configuration):
            result = {}
            for v in values[::-1]:
                if not isinstance(v, Configuration):
                    continue
                result.update(v)
            return Configuration(result)
        elif self._interpolate is not False:
            d = [d.as_dict() for d in self._configs]
            d[0].update(cast(Dict[str, str], self._interpolate))
            return interpolate_object(args[0], values[0], d, self._interpolate_type)
        else:
            return values[0]

    def _writable_config(self) -> Configuration:
        if not self._writable:
            lowercase = bool(self._configs and self._configs[0]._lowercase)
            self._configs.insert(0, Configuration({}, lowercase_keys=lowercase))
            self._writable = True
        return self._configs[0]

    @property
    def configs(self) -> List[Configuration]:
        """List of underlying configuration objects."""
        if self._writable:
            return self._configs[1:]
        else:
            return list(self._configs)

    @configs.setter
    def configs(self, iterable: Iterable[Configuration]) -> None:
        if self._writable:
            self._configs = [self._configs[0]] + list(iterable)
        else:
            self._configs = list(iterable)

    def __getitem__(self, item: str) -> Union[Configuration, Any]:  # noqa: D105
        return self._from_configs("__getitem__", item)

    def __getattr__(self, item: str) -> Union[Configuration, Any]:  # noqa: D105
        return self._from_configs("__getattr__", item)

    def get(self, key: str, default: Any = None) -> Union[dict, Any]:
        """
        Get the configuration values corresponding to :attr:`key`.

        :param key: key to retrieve
        :param default: default value in case the key is missing
        :return: the value found or a default
        """
        try:
            return self[key]
        except Exception:
            return default

    def as_dict(self) -> dict:
        """Return the representation as a dictionary."""
        result = {}
        for config_ in self._configs[::-1]:
            result.update(config_.as_dict())
        return result

    def get_dict(self, item: str) -> dict:
        """
        Get the item values as a dictionary.

        :param item: key
        """
        return Configuration({k: v for k, v in dict(self[item]).items()}).as_dict()

    def keys(
        self, levels: Optional[int] = None
    ) -> Union["Configuration", Any, KeysView[str]]:
        """Return a set-like object providing a view on the configuration keys."""
        if self._default_levels:
            return Configuration(self.as_dict()).keys(levels or self._default_levels)
        with Configuration(self.as_dict()).dotted_iter() as cfg:
            return cfg.keys(levels)

    def values(
        self, levels: Optional[int] = None
    ) -> Union["Configuration", Any, ValuesView[Any]]:
        """Return a set-like object providing a view on the configuration values."""
        if self._default_levels:
            return Configuration(self.as_dict()).values(levels or self._default_levels)
        with Configuration(self.as_dict()).dotted_iter() as cfg:
            return cfg.values(levels)

    def items(
        self, levels: Optional[int] = None
    ) -> Union["Configuration", Any, ItemsView[str, Any]]:
        """Return a set-like object providing a view on the configuration items."""
        if self._default_levels:
            return Configuration(self.as_dict()).items(levels or self._default_levels)
        with Configuration(self.as_dict()).dotted_iter() as cfg:
            return cfg.items(levels)

    def __setitem__(self, key: str, value: Any) -> None:  # noqa: D105
        cfg = self._writable_config()
        cfg[key] = value

    def __delitem__(self, prefix: str) -> None:  # noqa: D105
        removed = False
        for cfg in self._configs:
            try:
                del cfg[prefix]
                removed = True
            except KeyError:
                continue
        if not removed:
            raise KeyError()

    def __contains__(self, prefix: str) -> bool:  # noqa: D105
        return any(prefix in cfg for cfg in self._configs)

    def clear(self) -> None:
        """Remove all items."""
        for cfg in self._configs:
            cfg.clear()

    def copy(self) -> "Configuration":
        """Return shallow copy."""
        return ConfigurationSet(*self._configs)

    def update(self, other: Mapping[str, Any]) -> None:
        """Update the ConfigurationSet with another Configuration object or Mapping."""
        cfg = self._writable_config()
        cfg.update(other)

    def reload(self) -> None:
        """Reload the underlying configuration instances."""
        for cfg in self._configs:
            try:
                cfg.reload()
            except NotImplementedError:
                pass

    def __repr__(self) -> str:  # noqa: D105
        return "<ConfigurationSet: %s>" % hex(id(self))

    def __str__(self) -> str:  # noqa: D105
        return str({k: clean(k, v) for k, v in sorted(self.as_dict().items())})



from typing import List, Optional, TYPE_CHECKING, Union, Set

from ._data_provider import DataProviderLayer, Response
from ._data_type import DataType
from ..._core.session import get_valid_session
from ..._tools import validate_endpoint_request_url_parameters
from ...usage_collection._filter_types import FilterType
from ...usage_collection._logger import get_usage_logger
from ...usage_collection._utils import ModuleName

if TYPE_CHECKING:
    from ..._core.session._session import Session
    from ._endpoint_data import RequestMethod
    from ._data_provider import Response


class Definition(DataProviderLayer):
    """
    This class defines a wrapper around the Data (request/response)
    delivery mechanism of the platform

    Parameters
    ----------
    url : str
        The Url presents the address of the endpoint
    method : RequestMethod, optional
        The RESTful Data service can support multiple methods
        when sending requests to a specified endpoint
    path_parameters : dict, optional
        Defines the path variables used to apply to the request
    query_parameters : dict, optional
        Defines the query variables used to apply to the request
    header_parameters : dict, optional
        Defines the header variables used to apply to the request
    body_parameters : dict, optional
        Defines the body variables used to apply to the request


    Examples
    --------
    >>> from refinitiv.data.delivery import endpoint_request
    >>> definition_endpoint = endpoint_request.Definition("/data/news/v1/analyze")
    """

    # Should not change even if class name is changed
    _USAGE_CLS_NAME = "EndpointDefinition"

    def __init__(
        self,
        url: str,
        method: Union["RequestMethod", str, None] = None,
        path_parameters: Optional[dict] = None,
        query_parameters: Optional[dict] = None,
        header_parameters: Optional[dict] = None,
        body_parameters: Union[dict, List[dict], None] = None,
    ):
        self.url = url
        self.method = method
        self.path_parameters = path_parameters
        self.query_parameters = query_parameters
        self.body_parameters = body_parameters
        self.header_parameters = header_parameters
        super().__init__(
            data_type=DataType.ENDPOINT,
            url=self.url,
            method=self.method,
            path_parameters=self.path_parameters,
            query_parameters=self.query_parameters,
            body_parameters=self.body_parameters,
            header_parameters=self.header_parameters,
        )

    def get_data(self, session: Optional["Session"] = None) -> "Response":
        """
        Returns a response from the API to the library

        Parameters
        ----------
        session : Session, optional
            The Session defines the source where you want to retrieve your data

        Returns
        -------
        Response

        Examples
        --------
        >>> from refinitiv.data.delivery import endpoint_request
        >>> definition_endpoint = endpoint_request.Definition("/data/news/v1/analyze")
        >>> definition_endpoint.get_data()
        """
        validate_endpoint_request_url_parameters(self.url, self.path_parameters)

        session = get_valid_session(session)
        self._log_usage(
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}."
            f"get_data",
            {FilterType.SYNC, FilterType.LAYER_DELIVERY, FilterType.REST},
        )
        response = self._provider.get_data(
            session,
            self.url,
            method=self.method,
            path_parameters=self.path_parameters,
            query_parameters=self.query_parameters,
            header_parameters=self.header_parameters,
            body_parameters=self.body_parameters,
        )
        return response

    async def get_data_async(self, session: Optional["Session"] = None) -> "Response":
        """
        Returns a response asynchronously from the API to the library

        Parameters
        ----------
        session : Session, optional
            The Session defines the source where you want to retrieve your data

        Returns
        -------
        Response

        Examples
        --------
        >>> from refinitiv.data.delivery import endpoint_request
        >>> definition_endpoint = endpoint_request.Definition("/data/news/v1/analyze")
        >>> await definition_endpoint.get_data_async()
        """
        validate_endpoint_request_url_parameters(self.url, self.path_parameters)

        session = get_valid_session(session)
        self._log_usage(
            f"{self.__class__.__module__}."
            f"{self.__class__.__qualname__}."
            f"get_data_async",
            {FilterType.ASYNC, FilterType.LAYER_DELIVERY, FilterType.REST},
        )
        response = await self._provider.get_data_async(
            session,
            self.url,
            method=self.method,
            path_parameters=self.path_parameters,
            query_parameters=self.query_parameters,
            header_parameters=self.header_parameters,
            body_parameters=self.body_parameters,
        )
        return response

    def _log_usage(self, name: str, filter_type: Set[FilterType]):
        get_usage_logger().log_func(
            name=f"{ModuleName.DELIVERY}.{self._USAGE_CLS_NAME}.{name}",
            func_path=f"{self.__class__.__module__}.{self.__class__.__qualname__}.{name}",
            kwargs=dict(
                url=self.url,
                method=self.method,
                path_parameters=self.path_parameters,
                query_parameters=self.query_parameters,
                header_parameters=self.header_parameters,
                body_parameters=self.body_parameters,
            ),
            desc=filter_type,
        )



"""Functionality specific to special SearchTemplates template access class

Takes care of public API parts and loading user templates from the config
"""
from typing import Union, List

from humps import depascalize

from refinitiv.data import get_config
from .embedded import (
    RICCategoryTemplate,
    UnderlyingRICToOptionTemplate,
    UnderlyingRICToFutureTemplate,
    RICToIssuerTemplate,
    OrganisationPermIDToUPTemplate,
)
from .search import DiscoverySearchTemplate
from .namespaces import Namespace
from .utils import generate_docstring


def depascalize_view(value: str) -> str:
    """Convert search View value to upper snakecase

    We need to have separate function for that because some enum View values
    does not follow the rules of pascal case
    """
    if "STIRs" in value:
        value = value.replace("STIRs", "Stirs")

    return depascalize(value).upper()


def template_from_config_data(name: str, data: dict, config_prefix: str):
    namespace = Namespace(
        builtin=BuiltinNamespace(),
        user=UserNamespace(config_prefix=config_prefix),
        _=Namespace(
            **{  # TODO: locals can use global namespaces too
                name: template_from_config_data(name, sub, config_prefix)
                for name, sub in data.get("locals", {}).items()
            }
        ),
    )
    template = DiscoverySearchTemplate(
        name,
        placeholders_defaults={
            name: attrs["default"]
            for name, attrs in data.get("parameters", {}).items()
            if "default" in attrs
        },
        optional_placeholders={
            name
            for name, attrs in data.get("parameters", {}).items()
            if attrs.get("optional", False)
        },
        ns=namespace,
        **depascalize(data.get("request_body", {})),
    )

    method_args = {}
    # Some placeholders may be only in string, but not in "parameters"
    for param in sorted(template._placeholder_names):
        method_args[param] = data.get("parameters", {}).get(param, {})

    # That's why we can get them all in one cycle with pass-through parameters
    # that is located in "parameters" config session, but not in template string
    for param, desc in data.get("parameters", {}).items():
        if param not in template._placeholder_names:
            method_args[param] = desc

    template.__doc__ = generate_docstring(
        description=data.get("description", ""),
        methods={"search": {"description": "", "args": method_args}},
    )

    return template


class BuiltinNamespace(Namespace):
    def __getitem__(self, name: str):
        attr = getattr(SearchTemplates, name)
        if not isinstance(attr, DiscoverySearchTemplate):
            raise AttributeError(f"Embedded search template named {name} is not found")
        return attr


class UserNamespace(Namespace):
    _blacklisted_keys = {"request_body"}

    def __init__(self, config_prefix):
        super().__init__()

        self._CONFIG_PREFIX = config_prefix

    def __iter__(self):
        return get_config().get(self._CONFIG_PREFIX, {}).keys().__iter__()

    def __getitem__(self, name: str) -> Union[DiscoverySearchTemplate, "UserNamespace"]:
        if name in self._blacklisted_keys:
            raise KeyError(f"'{name}' is a reserved key")
        config = get_config()
        key = f"{self._CONFIG_PREFIX}.{name}"
        if key not in config:
            raise KeyError(f"Template or Namespace '{name}' is not found in the config")

        data = config[key] or {}
        if "request_body" not in data:
            return UserNamespace(key)

        data = config[key].as_attrdict() if data else {}
        return template_from_config_data(name, data, self._CONFIG_PREFIX)

    def keys(self) -> List[str]:
        """Get list of available search template names"""
        return list(self)


class SearchTemplates(UserNamespace):
    """Easy access to search templates from the library config

    Check if search template with the name "Equity" is defined in the config:
    >>> templates = SearchTemplates()
    >>> "Equity" in templates
    True
    Get "Equity" search template:
    >>> templates["Equity"]
    Get list of available search template names:
    >>> templates.keys()
    ["Equity"]
    """

    RICCategory = RICCategoryTemplate()
    UnderlyingRICToOption = UnderlyingRICToOptionTemplate()
    UnderlyingRICToFuture = UnderlyingRICToFutureTemplate()
    RICToIssuer = RICToIssuerTemplate()
    OrganisationPermIDToUP = OrganisationPermIDToUPTemplate()

    def __init__(self):
        super().__init__(config_prefix="search.templates")



import datetime

import pytest

from refinitiv.data._tools._datetime import (
    hp_datetime_adapter,
    cfs_datetime_adapter,
    tds_datetime_adapter,
    NanosecondsFormatter,
    ownership_datetime_adapter,
)


@pytest.fixture(scope="function", params=[hp_datetime_adapter, tds_datetime_adapter])
def zbase_datetime_adapter(request):
    adapter = request.param
    return adapter


@pytest.mark.parametrize(
    "input_value, expected",
    [
        ("12.02.2021", "2021-12-02T00:00:00.000000000Z"),
        ("12/02/2021", "2021-12-02T00:00:00.000000000Z"),
        (datetime.date(2021, 12, 2), "2021-12-02T00:00:00.000000000Z"),
    ],
)
def test_zbased_datetime_adapter_get_str(input_value, expected, zbase_datetime_adapter):
    # when
    testing_value = zbase_datetime_adapter.get_str(input_value)

    # then
    assert testing_value == expected, testing_value


def test_zbased_datetime_adapter_get_str_will_raise_error_is_pass_none(
    zbase_datetime_adapter,
):
    # given
    input_value = None

    # then
    with pytest.raises(AttributeError):
        # when
        zbase_datetime_adapter.get_str(input_value)


@pytest.mark.parametrize(
    "input_value, expected",
    (
        ("11.10.2020 15:00", "2020-11-10T15:00:00Z"),
        ("11.10.2020", "2020-11-10T00:00:00Z"),
        ("11.10.2020 15:21:01", "2020-11-10T15:21:01Z"),
        ("11/10/2020 15:21:01", "2020-11-10T15:21:01Z"),
        ("11/10/2020", "2020-11-10T00:00:00Z"),
        ("11-10-2020", "2020-11-10T00:00:00Z"),
    ),
)
def test_cfs_datetime_adapter_get_str(input_value, expected):
    # when
    testing_value = cfs_datetime_adapter.get_str(input_value)

    # then
    assert testing_value == expected


@pytest.mark.parametrize(
    "input_value, expected_value",
    [
        (datetime.timedelta(0), NanosecondsFormatter.NANOSECOND_LENGTH),
        (datetime.timedelta(-2), NanosecondsFormatter.NANOSECOND_LENGTH),
        (datetime.timedelta(2), NanosecondsFormatter.NANOSECOND_LENGTH),
    ],
)
def test_hp_datetime_adapter_return_correct_nanoseconds_with_timedelta(
    input_value, expected_value
):
    # when
    testing_value = hp_datetime_adapter.get_str(input_value)
    *_, nanoseconds_with_z = testing_value.split(".", maxsplit=1)
    nanoseconds = nanoseconds_with_z.replace("Z", "")
    testing_value = len(nanoseconds)

    # then
    assert testing_value == expected_value


@pytest.mark.parametrize(
    ("input_value", "expected_value"),
    [
        ("20211210", "20211210"),
        ("2021.12.10", "20211210"),
        ("-1QA", "-1QA"),
    ],
)
def test_ownership_datetime_adapter(input_value, expected_value):
    # given
    # when
    result = ownership_datetime_adapter.get_str(input_value)

    # then
    assert result == expected_value



# coding: utf8


from enum import Enum, unique


@unique
class CalibrationStrategy(Enum):
    ALL_MATURITY = "AllMaturity"
    DEFAULT = "Default"
    MGB_CALIBRATION = "MGBCalibration"
    MATURITY_BY_MATURITY = "MaturityByMaturity"



__all__ = (
    "Response",
    "cap",
    "eti",
    "fx",
    "swaption",
    "Definitions",
    "Outputs",
)

from . import cap
from . import eti
from . import fx
from . import swaption
from ....delivery._data._data_provider import Response
from ._definition import Definitions
from .._surfaces._enums import SurfaceOutputs as Outputs



from refinitiv.data.content.ipa.surfaces import swaption


def surface_swaption_definition():
    return swaption.Definition(
        surface_tag="My EUR VolCube",
        underlying_definition=swaption.SwaptionSurfaceDefinition(
            instrument_code="EUR",
            discounting_type=swaption.DiscountingType.OIS_DISCOUNTING,
        ),
        surface_parameters=swaption.SwaptionCalculationParams(
            shift_percent=3,
            x_axis=swaption.Axis.STRIKE,
            y_axis=swaption.Axis.TENOR,
            z_axis=swaption.Axis.EXPIRY,
        ),
        surface_layout=swaption.SurfaceLayout(
            format=swaption.Format.N_DIMENSIONAL_ARRAY,
        ),
    )


def invalid_surface_swaption_definition():
    return swaption.Definition(
        surface_tag="My EUR VolCube",
        underlying_definition=swaption.SwaptionSurfaceDefinition(
            instrument_code="INVAL",
            discounting_type=swaption.DiscountingType.OIS_DISCOUNTING,
        ),
        surface_parameters=swaption.SwaptionCalculationParams(
            shift_percent=3,
            x_axis=swaption.Axis.STRIKE,
            y_axis=swaption.Axis.TENOR,
            z_axis=swaption.Axis.EXPIRY,
        ),
        surface_layout=swaption.SurfaceLayout(
            format=swaption.Format.N_DIMENSIONAL_ARRAY,
        ),
    )



from copy import deepcopy
from datetime import datetime, timedelta, date

import pytest

import refinitiv.data as rd
from refinitiv.data._errors import RDError
from tests.unit.conftest import StubSession, StubResponse
from tests.unit.content.test_datetime import make_checker
from tests.unit.content.test_enums import get_json


def test_basic_case():
    # fmt: off
    expected_str = (
"  tenor   startDate     endDate  swapCcy1Bid  swapCcy1Ask  swapCcy2Bid  swapCcy2Ask  swapCc1Ccy2Bid  swapCc1Ccy2Ask  outrightCcy1Ccy2Bid  outrightCcy1Ccy2Ask\n"
"0    1M  2013-06-19  2013-07-19         1.84         1.88        -3.16        -3.11        2.852139        2.906775             0.850387             0.850908\n"
"1    2M  2013-06-19  2013-08-19         3.85         3.95        -6.25        -6.12        5.759443        5.897724             0.850678             0.851207"
    )
    raw = {
    "swapPoints": [{
        "tenor": "1M", "startDate": "2013-06-19",
        "endDate": "2013-07-19", "swapCcy1Bid": 1.84,
        "swapCcy1Ask": 1.88, "swapCcy2Bid": -3.16,
        "swapCcy2Ask": -3.11, "swapCc1Ccy2Bid": 2.85213907031734,
        "swapCc1Ccy2Ask": 2.90677481977641,
        "outrightCcy1Ccy2Bid": 0.850386969185587,
        "outrightCcy1Ccy2Ask": 0.850907804342901
    }, {
        "tenor": "2M", "startDate": "2013-06-19",
        "endDate": "2013-08-19", "swapCcy1Bid": 3.85,
        "swapCcy1Ask": 3.95, "swapCcy2Bid": -6.25,
        "swapCcy2Ask": -6.12, "swapCc1Ccy2Bid": 5.7594425853269,
        "swapCc1Ccy2Ask": 5.89772434189784,
        "outrightCcy1Ccy2Bid": 0.850677699537088,
        "outrightCcy1Ccy2Ask": 0.851206899295114
    }]
    }
    # fmt: on
    response = StubResponse(deepcopy(raw))
    session = StubSession(is_open=True, response=response)
    rd.session.set_default(session)
    response = rd._qpl.fx_swp_to_swp(
        fx_cross_code="EURGBP",
        market_data_date_time="2013-06-17",
        tenors=["1M", "2M"],
    )
    assert response.data.raw == raw
    assert response.data.df.to_string() == expected_str

    rd.session.set_default(None)


def test_manage_error_and_processing_information():
    error_code = "QPS-Curves.14"
    error_message = (
        "The service failed to build the curve, Invalid input : "
        "Definition of the fx forward curve is not valid. "
        "The fx forward curve doesn't contain a valid spot "
        "constituent for currency code 'XXXYYY'."
    )
    raw = {
        "error": {
            "id": "fe68b128-2ea6-44d7-9456-3196acd5856e/fe68b128-2ea6-44d7-9456-3196acd5856e",
            "status": "Error",
            "message": error_message,
            "code": error_code,
        }
    }
    response = StubResponse(deepcopy(raw))
    session = StubSession(is_open=True, response=response)
    rd.session.set_default(session)
    with pytest.raises(RDError) as error:
        rd._qpl.fx_swp_to_swp(
            fx_cross_code="XXXYYY",
            market_data_date_time="2013-06-17",
            tenors=["1M"],
        )

    assert error.value.code == error_code
    assert error.value.message == error_message
    rd.session.set_default(None)


def test_full_override_of_spot_and_swap_points_values():
    # fmt: off
    expected_str = (
"  tenor   startDate     endDate  swapCcy1Bid  swapCcy1Ask  swapCcy2Bid  swapCcy2Ask  swapCc1Ccy2Bid  swapCc1Ccy2Ask  outrightCcy1Ccy2Bid  outrightCcy1Ccy2Ask\n"
"0    1M  2022-11-21  2022-12-21        22.16        22.45         8.64         9.04       12.014555       12.552910             0.874284             0.874970\n"
"1    2M  2022-11-21  2023-01-23        55.66        56.16        24.49        28.49       25.882465       29.245585             0.875671             0.876639"
    )
    raw = {
        "swapPoints": [{
            "tenor": "1M", "startDate": "2022-11-21",
            "endDate": "2022-12-21", "swapCcy1Bid": 22.16,
            "swapCcy1Ask": 22.45, "swapCcy2Bid": 8.64,
            "swapCcy2Ask": 9.04, "swapCc1Ccy2Bid": 12.0145547910377,
            "swapCc1Ccy2Ask": 12.552910153959,
            "outrightCcy1Ccy2Bid": 0.874284212937388,
            "outrightCcy1Ccy2Ask": 0.874969673075757
        }, {
            "tenor": "2M", "startDate": "2022-11-21",
            "endDate": "2023-01-23", "swapCcy1Bid": 55.66,
            "swapCcy1Ask": 56.16, "swapCcy2Bid": 24.49,
            "swapCcy2Ask": 28.49, "swapCc1Ccy2Bid": 25.8824651078038,
            "swapCc1Ccy2Ask": 29.2455845109385,
            "outrightCcy1Ccy2Bid": 0.875671003969064,
            "outrightCcy1Ccy2Ask": 0.876638940511455
        }]
    }
    expected_raw = deepcopy(raw)
    expected_json = {
        'fxCrossCode': 'EURGBP', 'marketDataDate': '2022-11-17', 'tenors': ['1M', '2M'],
        'spotCcy1': {'ask': 1, 'bid': 1}, 'spotCcy2': {'ask': 1.3, 'bid': 1.3},
        'swapPointsCcy1': {
            'overrides': [{'tenor': '1M', 'ask': 30, 'bid': 20},
                          {'tenor': '2M', 'ask': 45, 'bid': 34}]
        },
        'swapPointsCcy2': {
            'overrides': [{'tenor': '1M', 'ask': 40, 'bid': 40},
                          {'tenor': '2M', 'ask': 45, 'bid': 45}]
        }
    }
    # fmt: on

    def make_assert_request():
        origin_session_http_request = session.http_request

        def assert_request(request):
            assert request.json == expected_json
            return origin_session_http_request(request)

        return assert_request

    response = StubResponse(raw)
    session = StubSession(is_open=True, response=response)
    session.http_request = make_assert_request()
    rd.session.set_default(session)
    response = rd._qpl.fx_swp_to_swp(
        fx_cross_code="EURGBP",
        market_data_date_time="2022-11-17",
        tenors=["1M", "2M"],
        spot_ccy_1=rd._qpl.FxSpotQuote(bid=1, ask=1),
        spot_ccy_2=rd._qpl.FxSpotQuote(bid=1.3, ask=1.3),
        swap_points_ccy_1=rd._qpl.FxSwapPoints(
            overrides=[
                rd._qpl.TenorBidAsk(tenor="1M", bid=20, ask=30),
                rd._qpl.TenorBidAsk(tenor="2M", bid=34, ask=45),
            ]
        ),
        swap_points_ccy_2=rd._qpl.FxSwapPoints(
            overrides=[
                rd._qpl.TenorBidAsk(tenor="1M", bid=40, ask=40),
                rd._qpl.TenorBidAsk(tenor="2M", bid=45, ask=45),
            ]
        ),
    )
    assert response.data.raw == expected_raw
    assert response.data.df.to_string() == expected_str

    rd.session.set_default(None)


def test_case_19():
    # fmt: off
    expected_str = (
"  tenor   startDate     endDate  swapCcy1Bid  swapCcy1Ask  swapCcy2Bid  swapCcy2Ask  swapCc1Ccy2Bid  swapCc1Ccy2Ask  outrightCcy1Ccy2Bid  outrightCcy1Ccy2Ask\n"
"0    1M  2022-11-21  2022-12-21        22.16        22.45         8.64         9.04       12.014555       12.552910             0.874284             0.874970\n"
"1    2M  2022-11-21  2023-01-23        55.66        56.16        24.49        28.49       25.882465       29.245585             0.875671             0.876639"
    )
    raw = {
        "swapPoints": [{
            "tenor": "1M", "startDate": "2022-11-21",
            "endDate": "2022-12-21", "swapCcy1Bid": 22.16,
            "swapCcy1Ask": 22.45, "swapCcy2Bid": 8.64,
            "swapCcy2Ask": 9.04, "swapCc1Ccy2Bid": 12.0145547910377,
            "swapCc1Ccy2Ask": 12.552910153959,
            "outrightCcy1Ccy2Bid": 0.874284212937388,
            "outrightCcy1Ccy2Ask": 0.874969673075757
        }, {
            "tenor": "2M", "startDate": "2022-11-21",
            "endDate": "2023-01-23", "swapCcy1Bid": 55.66,
            "swapCcy1Ask": 56.16, "swapCcy2Bid": 24.49,
            "swapCcy2Ask": 28.49, "swapCc1Ccy2Bid": 25.8824651078038,
            "swapCc1Ccy2Ask": 29.2455845109385,
            "outrightCcy1Ccy2Bid": 0.875671003969064,
            "outrightCcy1Ccy2Ask": 0.876638940511455
        }]
    }
    expected_raw = deepcopy(raw)
    expected_json = {
        'fxCrossCode': 'EURUSD',
        'marketDataDate': '2022-09-22',
        'spotCcy1': {'source': 'D3'},
        'spotCcy2': {'ask': 2, 'bid': 1},
        'swapPointsCcy1': {'additionalTenorTypes': ['Long'], 'source': 'ICAP'},
        'swapPointsCcy2': {
            'additionalTenorTypes': ['Long', 'Odd'],
            'overrides': [{'ask': 60, 'bid': 50, 'tenor': '1M'},
                          {'bid': 90, 'tenor': '2M'}],
            'source': 'D3'
        }
    }
    # fmt: on

    def make_assert_request():
        origin_session_http_request = session.http_request

        def assert_request(request):
            assert request.json == expected_json
            return origin_session_http_request(request)

        return assert_request

    response = StubResponse(raw)
    session = StubSession(is_open=True, response=response)
    session.http_request = make_assert_request()
    rd.session.set_default(session)
    response = rd._qpl.fx_swp_to_swp(
        fx_cross_code="EURUSD",
        market_data_date_time="2022-09-22",
        spot_ccy_1=rd._qpl.FxSpotQuote(source="D3"),
        spot_ccy_2=rd._qpl.FxSpotQuote(bid=1, ask=2),
        swap_points_ccy_1=rd._qpl.FxSwapPoints(
            additional_tenor_types=[rd._qpl.TenorTypes.LONG],
            source="ICAP",
        ),
        swap_points_ccy_2=rd._qpl.FxSwapPoints(
            additional_tenor_types=[rd._qpl.TenorTypes.LONG, rd._qpl.TenorTypes.ODD],
            source="D3",
            overrides=[
                rd._qpl.TenorBidAsk(tenor="1M", bid=50, ask=60),
                rd._qpl.TenorBidAsk(tenor="2M", bid=90),
            ],
        ),
    )
    assert response.data.raw == expected_raw
    assert response.data.df.to_string() == expected_str

    rd.session.set_default(None)


@pytest.mark.parametrize(
    (
        "input_datatime",
        "expected_str",
    ),
    [
        (datetime(2020, 1, 1), "2020-01-01"),
        ("2020-01-02", "2020-01-02"),
        (timedelta(0), datetime.now().date()),
        (date(2020, 1, 4), "2020-01-04"),
    ],
)
def test_qpl_function_datetime(input_datatime, expected_str):
    session = StubSession(is_open=True)
    session.http_request = make_checker(
        param_name="marketDataDate",
        value=expected_str,
        func_get_params=get_json,
        mocked_json=None,
    )
    rd.session.set_default(session)

    rd._qpl.fx_swp_to_swp(fx_cross_code="code", market_data_date_time=input_datatime)

    rd.session.set_default(None)



import pytest

from refinitiv.data.content.ipa import _surfaces
from refinitiv.data.content.ipa._surfaces._models import Surface
from refinitiv.data.content.ipa._surfaces._surfaces_data_provider import (
    OneSurfaceData,
    SurfacesData,
    BaseData,
    SurfaceResponseFactory,
)
from refinitiv.data.content.ipa.surfaces import eti
from refinitiv.data.delivery._data._data_provider import ParsedData
from tests.unit.content.ipa._surfaces.conftest import (
    one_surface_data_json,
    surfaces_data_json,
    one_surface_axis_params,
    surfaces_axis_params,
)


def test_surfaces_version():
    # given
    expected = "1.0.59"

    # when
    testing_version = _surfaces.__version__

    # then
    assert testing_version == expected, testing_version


def test_one_surface_data_has_one_surface():
    # given
    data = OneSurfaceData(one_surface_data_json, one_surface_axis_params)

    # when
    surface = data.surface

    # then
    assert isinstance(surface, Surface) is True


def test_one_surface_data_does_not_have_surfaces():
    # given
    data = OneSurfaceData(one_surface_data_json, one_surface_axis_params)

    # then
    with pytest.raises(AttributeError):
        # when
        data.surfaces


def test_surfaces_data_has_many_surfaces():
    # given
    data = SurfacesData(surfaces_data_json, surfaces_axis_params)

    # when
    surfaces = data.surfaces

    # then
    assert len(surfaces) > 1


def test_one_surface_data_does_not_have_surface():
    # given
    data = SurfacesData(surfaces_data_json, surfaces_axis_params)

    # then
    with pytest.raises(AttributeError):
        # when
        data.surface


def test_base_data_has_df_property():
    # given
    data = BaseData(one_surface_data_json)

    # when
    df = data.df

    # then
    assert not df.empty


def test_base_data_has_df_if_data_raw_is_empty():
    # given
    data = BaseData({"data": {}})

    # when
    df = data.df

    # then
    assert df.empty is True


def test_base_data_does_not_have_surface_property():
    # given
    data = BaseData(one_surface_data_json)

    # then
    with pytest.raises(AttributeError):
        # when
        data.surface


def test_base_data_does_not_have_surfaces_property():
    # given
    data = BaseData(one_surface_data_json)

    # then
    with pytest.raises(AttributeError):
        # when
        data.surfaces


def test_surface_response_factory_can_create_response_from_one_universe():
    # given
    definition = eti.Definition(
        underlying_definition=eti.EtiSurfaceDefinition(instrument_code="BNPP.PA@RIC"),
        surface_parameters=eti.EtiCalculationParams(
            price_side=eti.PriceSide.MID,
            volatility_model=eti.VolatilityModel.SVI,
            x_axis=eti.Axis.STRIKE,
            y_axis=eti.Axis.DATE,
        ),
        surface_tag="1_MATRIX",
        surface_layout=eti.SurfaceLayout(format=eti.Format.MATRIX, y_point_count=10),
    )
    response_factory = SurfaceResponseFactory()

    # when
    response = response_factory.create_success(
        ParsedData({}, {}, one_surface_data_json),
        universe=definition,
    )

    # then
    assert isinstance(response.data.surface, Surface) is True


def test_surface_response_factory_can_create_response_from_many_universes():
    # given
    universe = [
        eti.Definition(
            underlying_definition=eti.EtiSurfaceDefinition(
                instrument_code="BNPP.PA@RIC"
            ),
            surface_parameters=eti.EtiCalculationParams(
                price_side=eti.PriceSide.MID,
                volatility_model=eti.VolatilityModel.SVI,
                x_axis=eti.Axis.STRIKE,
                y_axis=eti.Axis.DATE,
            ),
            surface_tag="1_MATRIX",
            surface_layout=eti.SurfaceLayout(
                format=eti.Format.MATRIX, y_point_count=10
            ),
        ),
        eti.Definition(
            underlying_definition=eti.EtiSurfaceDefinition(
                instrument_code="BNPP.PA@RIC"
            ),
            surface_parameters=eti.EtiCalculationParams(
                price_side=eti.PriceSide.MID,
                volatility_model=eti.VolatilityModel.SVI,
                x_axis=eti.Axis.DATE,
                y_axis=eti.Axis.STRIKE,
            ),
            surface_tag="1_1",
            surface_layout=eti.SurfaceLayout(
                format=eti.Format.MATRIX, y_point_count=10
            ),
        ),
    ]
    response_factory = SurfaceResponseFactory()

    # when
    response = response_factory.create_success(
        ParsedData({}, {}, surfaces_data_json),
        universe=universe,
    )

    # then
    assert len(response.data.surfaces) > 1


def test_surface_response_factory_can_create_response_with_base_data():
    # given
    response_factory = SurfaceResponseFactory()

    # when
    response = response_factory.create_success(
        ParsedData({}, {}, surfaces_data_json),
    )
    data = response.data

    # then
    assert isinstance(data, BaseData)



import allure
import pytest

from refinitiv.data.content.ownership import StatTypes
from refinitiv.data.content.ownership.consolidated import breakdown
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.ownership.helpers import (
    check_response_dataframe_instrument_contains_expected_universes,
    get_expected_column,
)
from tests.integration.helpers import (
    get_async_response_from_definitions,
    get_async_response_from_definition,
    check_response_status,
    check_non_empty_response_data,
    check_response_dataframe_contains_columns_names,
    check_extended_params_were_sent_in_request,
)

EXPECTED_COLUMN_NAMES = [
    "instrument",
    "categoryvalue",
    "TR.CategoryInvestorCount",
    "TR.CategoryOwnershipPct",
    "TR.InstrStatCatSharesHeld",
    "TR.InstrStatCatShrsHldVal",
]

EXPECTED_COLUMN_TITLES = [
    "Instrument",
    "Category Value",
    "Category Investor Count",
    "Category Ownership %",
    "Category SharesHeld",
    "Category SharesHeld Value",
]


@allure.suite("Content object - Ownership Consolidated Breakdown")
@allure.feature("Content object - Ownership Consolidated Breakdown")
@allure.severity(allure.severity_level.CRITICAL)
class TestConsolidatedBreakdown:
    @allure.title(
        "Create Consolidated Breakdown definition object with valid required params and get data"
    )
    @pytest.mark.caseid("37556928")
    @pytest.mark.parametrize(
        "universe,stat_type,use_field_names_in_headers,extended_params,expected_universe",
        [
            ("ONEX.CCP", 1, True, {"test_param": "test_value"}, "ONEX.CCP"),
            ("US02079K1079", 3, False, None, "US02079K1079"),
            (
                ["TRI.N", "ONEX.CCP", "AAPL.O"],
                StatTypes.REGION,
                False,
                {"universe": "ONEX.CCP,US02079K1079"},
                ["ONEX.CCP", "US02079K1079"],
            ),
            (
                ["ONEX.CCP", "invalid universe"],
                StatTypes.COUNTRY,
                True,
                None,
                ["ONEX.CCP", "invalid universe"],
            ),
        ],
    )
    def test_create_consolidated_breakdown_object_with_valid_fields_and_get_data(
        self,
        open_platform_session,
        universe,
        stat_type,
        use_field_names_in_headers,
        extended_params,
        expected_universe,
    ):
        expected_column = get_expected_column(
            use_field_names_in_headers, EXPECTED_COLUMN_TITLES, EXPECTED_COLUMN_NAMES
        )
        response = breakdown.Definition(
            universe, stat_type, use_field_names_in_headers, extended_params
        ).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_response_dataframe_contains_columns_names(response, expected_column)
        check_response_dataframe_instrument_contains_expected_universes(
            response, expected_universe
        )
        check_extended_params_were_sent_in_request(response, extended_params)

    @allure.title(
        "Create Consolidated Breakdown definition object with mix valid and invalid params and get data asynchronously"
    )
    @pytest.mark.caseid("37556929")
    @pytest.mark.parametrize(
        "universe,invalid_universe,stat_type",
        [
            (["ONEX.CCP", "invalid universe"], "invalid universe", StatTypes.COUNTRY),
        ],
    )
    @pytest.mark.smoke
    async def test_create_consolidated_breakdown_object_with_mix_valid_and_invalid_params_and_get_data_async(
        self, open_platform_session_async, universe, invalid_universe, stat_type
    ):
        valid_response, invalid_response = await get_async_response_from_definitions(
            breakdown.Definition(universe, stat_type, use_field_names_in_headers=True),
            breakdown.Definition(
                invalid_universe, stat_type, use_field_names_in_headers=True
            ),
        )
        check_response_status(
            response=valid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code=None,
            expected_error_message="Failed to resolve identifiers ['invalid universe']",
        )
        check_non_empty_response_data(valid_response)
        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code=HttpStatusCode.FOUR_HUNDRED_TWELVE,
            expected_error_message="Unable to resolve all requested identifiers. Requested items: ['invalid universe']",
        )
        check_response_dataframe_contains_columns_names(
            valid_response, EXPECTED_COLUMN_NAMES
        )
        check_response_dataframe_instrument_contains_expected_universes(
            valid_response, universe
        )

    @allure.title(
        "Create Consolidated Breakdown definition object with invalid fields and get error"
    )
    @pytest.mark.caseid("37556930")
    @pytest.mark.parametrize(
        "universe,stat_type",
        [
            ([], 2),
            ([""], 3),
        ],
    )
    def test_create_consolidated_breakdown_object_with_invalid_fields_and_get_error(
        self, open_platform_session, universe, stat_type
    ):
        with pytest.raises(RDError) as error:
            breakdown.Definition(universe, stat_type).get_data()
        assert (
            str(error.value)
            == "Error code 400 | Validation error: validation failure list:\nuniverse.0 in query should be at least 1 chars long"
        )

    @allure.title("Create Consolidated Breakdown definition object with closed session")
    @pytest.mark.caseid("37556931")
    async def test_create_consolidated_breakdown_object_with_closed_session(
        self, open_platform_session_async
    ):
        session = open_platform_session_async
        await session.close_async()
        definition = breakdown.Definition("ONEX.CCP", 1)
        with pytest.raises(ValueError) as error:
            await get_async_response_from_definition(definition)
        assert str(error.value) == "Session is not opened. Can't send any request"



# coding: utf8
from .._object_definition import ObjectDefinition

from ._enums import FxVolatilityModel
from ._enums import FxSwapCalculationMethod
from ._enums import PriceSide
from ._enums import TimeStamp
from ._enums import Axis
from ._models import BidAskMid
from ._models import InterpolationWeight


class FxSurfaceParameters(ObjectDefinition):
    def __init__(
        self,
        atm_volatility_object=None,
        butterfly10_d_object=None,
        butterfly25_d_object=None,
        domestic_deposit_rate_percent_object=None,
        foreign_deposit_rate_percent_object=None,
        forward_points_object=None,
        fx_spot_object=None,
        fx_swap_calculation_method=None,
        implied_volatility_object=None,
        interpolation_weight=None,
        price_side=None,
        risk_reversal10_d_object=None,
        risk_reversal25_d_object=None,
        time_stamp=None,
        volatility_model=None,
        x_axis=None,
        y_axis=None,
        calculation_date=None,
    ):
        super().__init__()
        self.atm_volatility_object = atm_volatility_object
        self.butterfly10_d_object = butterfly10_d_object
        self.butterfly25_d_object = butterfly25_d_object
        self.domestic_deposit_rate_percent_object = domestic_deposit_rate_percent_object
        self.foreign_deposit_rate_percent_object = foreign_deposit_rate_percent_object
        self.forward_points_object = forward_points_object
        self.fx_spot_object = fx_spot_object
        self.fx_swap_calculation_method = fx_swap_calculation_method
        self.implied_volatility_object = implied_volatility_object
        self.interpolation_weight = interpolation_weight
        self.price_side = price_side
        self.risk_reversal10_d_object = risk_reversal10_d_object
        self.risk_reversal25_d_object = risk_reversal25_d_object
        self.time_stamp = time_stamp
        self.volatility_model = volatility_model
        self.x_axis = x_axis
        self.y_axis = y_axis
        self.calculation_date = calculation_date

    @property
    def atm_volatility_object(self):
        """
        At the money volatility at Expiry
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "atmVolatilityObject")

    @atm_volatility_object.setter
    def atm_volatility_object(self, value):
        self._set_object_parameter(BidAskMid, "atmVolatilityObject", value)

    @property
    def butterfly10_d_object(self):
        """
        BF 10 Days at Expiry
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "butterfly10DObject")

    @butterfly10_d_object.setter
    def butterfly10_d_object(self, value):
        self._set_object_parameter(BidAskMid, "butterfly10DObject", value)

    @property
    def butterfly25_d_object(self):
        """
        BF 25 Days at Expiry
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "butterfly25DObject")

    @butterfly25_d_object.setter
    def butterfly25_d_object(self, value):
        self._set_object_parameter(BidAskMid, "butterfly25DObject", value)

    @property
    def domestic_deposit_rate_percent_object(self):
        """
        Domestic Deposit Rate at Expiry
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "domesticDepositRatePercentObject")

    @domestic_deposit_rate_percent_object.setter
    def domestic_deposit_rate_percent_object(self, value):
        self._set_object_parameter(BidAskMid, "domesticDepositRatePercentObject", value)

    @property
    def foreign_deposit_rate_percent_object(self):
        """
        Foreign Deposit Rate at Expiry
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "foreignDepositRatePercentObject")

    @foreign_deposit_rate_percent_object.setter
    def foreign_deposit_rate_percent_object(self, value):
        self._set_object_parameter(BidAskMid, "foreignDepositRatePercentObject", value)

    @property
    def forward_points_object(self):
        """
        Forward Points at Expiry
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "forwardPointsObject")

    @forward_points_object.setter
    def forward_points_object(self, value):
        self._set_object_parameter(BidAskMid, "forwardPointsObject", value)

    @property
    def fx_spot_object(self):
        """
        Spot Price
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "fxSpotObject")

    @fx_spot_object.setter
    def fx_spot_object(self, value):
        self._set_object_parameter(BidAskMid, "fxSpotObject", value)

    @property
    def fx_swap_calculation_method(self):
        """
        The method we chose to price outrights using or not implied deposits. Possible values are:

         FxSwap (compute outrights using swap points),

         DepositCcy1ImpliedFromFxSwap (compute currency1 deposits using swap points),

         DepositCcy2ImpliedFromFxSwap (compute currency2 deposits using swap points).

         Optional. Defaults to 'FxSwap'.
        :return: enum FxSwapCalculationMethod
        """
        return self._get_enum_parameter(
            FxSwapCalculationMethod, "fxSwapCalculationMethod"
        )

    @fx_swap_calculation_method.setter
    def fx_swap_calculation_method(self, value):
        self._set_enum_parameter(
            FxSwapCalculationMethod, "fxSwapCalculationMethod", value
        )

    @property
    def implied_volatility_object(self):
        """
        Implied Volatility at Expiry
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "impliedVolatilityObject")

    @implied_volatility_object.setter
    def implied_volatility_object(self, value):
        self._set_object_parameter(BidAskMid, "impliedVolatilityObject", value)

    @property
    def interpolation_weight(self):
        """
        Vol Term Structure Interpolation
        :return: object InterpolationWeight
        """
        return self._get_object_parameter(InterpolationWeight, "interpolationWeight")

    @interpolation_weight.setter
    def interpolation_weight(self, value):
        self._set_object_parameter(InterpolationWeight, "interpolationWeight", value)

    @property
    def price_side(self):
        """
        Specifies whether bid, ask or mid is used to build the surface.
        :return: enum PriceSide
        """
        return self._get_enum_parameter(PriceSide, "priceSide")

    @price_side.setter
    def price_side(self, value):
        self._set_enum_parameter(PriceSide, "priceSide", value)

    @property
    def risk_reversal10_d_object(self):
        """
        RR 10 Days at Expiry
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "riskReversal10DObject")

    @risk_reversal10_d_object.setter
    def risk_reversal10_d_object(self, value):
        self._set_object_parameter(BidAskMid, "riskReversal10DObject", value)

    @property
    def risk_reversal25_d_object(self):
        """
        RR 25 Days at Expiry
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "riskReversal25DObject")

    @risk_reversal25_d_object.setter
    def risk_reversal25_d_object(self, value):
        self._set_object_parameter(BidAskMid, "riskReversal25DObject", value)

    @property
    def time_stamp(self):
        """
        Define how the timestamp is selected:
        - Open: the opening value of the valuationDate or if not available the close of the previous day is used.
        - Default: the latest snapshot is used when valuationDate is today, the close price when valuationDate is in the past.
        :return: enum TimeStamp
        """
        return self._get_enum_parameter(TimeStamp, "timeStamp")

    @time_stamp.setter
    def time_stamp(self, value):
        self._set_enum_parameter(TimeStamp, "timeStamp", value)

    @property
    def volatility_model(self):
        """
        The quantitative model used to generate the volatility surface. This may depend on the asset class.
        For Fx Volatility Surface, we currently support the SVI model.
        :return: enum FxVolatilityModel
        """
        return self._get_enum_parameter(FxVolatilityModel, "volatilityModel")

    @volatility_model.setter
    def volatility_model(self, value):
        self._set_enum_parameter(FxVolatilityModel, "volatilityModel", value)

    @property
    def x_axis(self):
        """
        Specifies the unit for the x axis (e.g. Date, Tenor)
        :return: enum Axis
        """
        return self._get_enum_parameter(Axis, "xAxis")

    @x_axis.setter
    def x_axis(self, value):
        self._set_enum_parameter(Axis, "xAxis", value)

    @property
    def y_axis(self):
        """
        Specifies the unit for the y axis (e.g. Strike, Delta). This may depend on the asset class.
        For Fx Volatility Surface, we support both Delta and Strike.
        :return: enum Axis
        """
        return self._get_enum_parameter(Axis, "yAxis")

    @y_axis.setter
    def y_axis(self, value):
        self._set_enum_parameter(Axis, "yAxis", value)

    @property
    def calculation_date(self):
        """
        The date the volatility surface is generated.
        :return: str
        """
        return self._get_parameter("calculationDate")

    @calculation_date.setter
    def calculation_date(self, value):
        self._set_parameter("calculationDate", value)



forward_curve_universe = {
    "curveDefinition": {
        "indexName": "EURIBOR",
        "currency": "EUR",
        "discountingTenor": "OIS",
        "name": "EUR EURIBOR Swap ZC Curve",
    },
    "curveParameters": {"calendarAdjustment": "Calendar"},
    "forwardCurveDefinitions": [
        {
            "indexTenor": "3M",
            "forwardCurveTenors": ["0D", "1D"],
            "forwardCurveTag": "MyForwardTag",
            "forwardStartDate": "2021-02-01",
            "forwardStartTenor": "some_start_tenor",
        }
    ],
    "curveTag": "my_test_curve",
}



# coding: utf8


from enum import Enum, unique


@unique
class DividendExtrapolation(Enum):
    CST_EXTRAPOL = "Cst_Extrapol"
    LINEAR_EXTRAPOL = "Linear_Extrapol"
    POWER_GROWTH_EXTRAPOL = "PowerGrowth_Extrapol"



from refinitiv.data.content.ipa._curves._cross_currency_curves._definitions import (
    FieldFormulaDescription,
)


def test_field_formula_description():
    # given
    expected_dict = {
        "historicalFidPriority": ["string1", "string2"],
        "realTimeFidPriority": ["string1", "string2"],
        "formula": "string",
    }

    # when
    testing_obj = FieldFormulaDescription(
        historical_fid_priority=["string1", "string2"],
        real_time_fid_priority=["string1", "string2"],
        formula="string",
    )

    # then
    assert testing_obj.get_dict() == expected_dict

    assert testing_obj.historical_fid_priority == ["string1", "string2"]
    assert testing_obj.real_time_fid_priority == ["string1", "string2"]
    assert testing_obj.formula == "string"



__all__ = (
    "AdjustInterestToPaymentDate",
    "AmortizationFrequency",
    "AmortizationItem",
    "AmortizationType",
    "BarrierDefinitionElement",
    "BarrierType",
    "BusinessDayConvention",
    "BuySell",
    "DateRollingConvention",
    "DayCountBasis",
    "Definition",
    "Frequency",
    "IndexConvexityAdjustmentIntegrationMethod",
    "IndexConvexityAdjustmentMethod",
    "IndexResetType",
    "InputFlow",
    "InterestCalculationConvention",
    "PremiumSettlementType",
    "PriceSide",
    "PricingParameters",
    "StubRule",
)

from ._cap_floor_pricing_parameters import PricingParameters
from ._definition import Definition
from ._enums import (
    AdjustInterestToPaymentDate,
    AmortizationFrequency,
    AmortizationType,
    BarrierType,
    BusinessDayConvention,
    BuySell,
    DateRollingConvention,
    DayCountBasis,
    Frequency,
    IndexConvexityAdjustmentIntegrationMethod,
    IndexConvexityAdjustmentMethod,
    IndexResetType,
    InterestCalculationConvention,
    PremiumSettlementType,
    PriceSide,
    StubRule,
)

from ._models import AmortizationItem, BarrierDefinitionElement, InputFlow



import asyncio
import json
import os
import socket
from functools import partial
from threading import Event
from typing import Callable, Optional, TYPE_CHECKING, Any, Union

from ._protocol_type import ProtocolType
from ._stream_listener import OMMStreamListener, StreamEvent
from .contrib import ContribResponse
from .contrib import contrib_type_enum_arg_parser
from .contrib._response import (
    ErrorContribResponse,
    AckContribResponse,
    NullContribResponse,
)
from .stream import (
    Stream,
    update_message_with_extended_params,
)
from ..._tools import cached_property, OrEvent

if TYPE_CHECKING:
    from .contrib import ContribType
    from ..._types import OptStr, ExtendedParams, Strings
    from ._stream_factory import StreamDetails
    from . import StreamConnection
    from ..._core.session import Session


class _OMMStream(Stream, OMMStreamListener["_OMMStream"]):
    def __init__(
        self,
        stream_id: int,
        session: "Session",
        name: str,
        details: "StreamDetails",
        domain: "OptStr" = "MarketPrice",
        service: "OptStr" = None,
        fields: Optional["Strings"] = None,
        key: Optional[dict] = None,
        extended_params: "ExtendedParams" = None,
        on_refresh: Optional[Callable] = None,
        on_status: Optional[Callable] = None,
        on_update: Optional[Callable] = None,
        on_complete: Optional[Callable] = None,
        on_error: Optional[Callable] = None,
        on_ack: Optional[Callable] = None,
    ) -> None:
        Stream.__init__(self, stream_id, session, details)
        OMMStreamListener.__init__(
            self,
            logger=session.logger(),
            on_refresh=on_refresh,
            on_status=on_status,
            on_update=on_update,
            on_complete=on_complete,
            on_error=on_error,
            on_ack=on_ack,
        )
        self._name = name
        self._service = service
        self._fields = fields or []
        self._domain = domain
        self._key = key
        self._extended_params = extended_params

        self._with_updates: bool = True
        self._refresh_message = None
        self._status_message = None
        self._error_message = None
        self._ack_message = None

        self._error_event = Event()
        self._ack_event = Event()
        self._update_event = Event()

        self._contrib_response = None
        self._is_contributing = False
        self._post_id = None
        self._post_user_info = None

    @cached_property
    def _contributed(self) -> OrEvent:
        return OrEvent(self._error_event, self._update_event)

    @property
    def post_id(self) -> int:
        return self._post_id

    def get_next_post_id(self) -> int:
        return next(self.session._contrib_post_id_counter)

    @property
    def post_user_info(self) -> dict:
        return self._post_user_info

    @property
    def name(self) -> str:
        return self._name

    @property
    def protocol_type(self) -> ProtocolType:
        return ProtocolType.OMM

    @property
    def domain(self) -> str:
        return self._domain

    @property
    def stream_state(self) -> str:
        return self._stream_state

    @property
    def message_state(self) -> str:
        return self._message_state

    @property
    def service(self):
        return self._service

    @property
    def is_contributing(self):
        return self._is_contributing

    @property
    def open_message(self):
        msg = {
            "ID": self.id,
            "Domain": self._domain,
            "Streaming": self._with_updates,
            "Key": {},
        }

        if self._key:
            msg["Key"] = self._key

        msg["Key"]["Name"] = self.name

        if self.service:
            msg["Key"]["Service"] = self.service

        if self._fields:
            msg["View"] = self._fields

        if self._extended_params:
            msg = update_message_with_extended_params(msg, self._extended_params)

        return msg

    @cached_property
    def close_message(self):
        return {"ID": self.id, "Type": "Close"}

    def _do_open(self, *args, **kwargs):
        self._with_updates = kwargs.get("with_updates", True)
        self._initialize_cxn()
        self._cxn.on(self._event.refresh_by_id, self._on_stream_refresh)
        self._cxn.on(self._event.update_by_id, self._on_stream_update)
        self._cxn.on(self._event.status_by_id, self._on_stream_status)
        self._cxn.on(self._event.complete_by_id, self._on_stream_complete)
        self._cxn.on(self._event.error_by_id, self._on_stream_error)
        self._cxn.on(self._event.ack_by_id, self._on_stream_ack)
        super()._do_open(*args, **kwargs)

    def _dispose(self):
        self._debug(f"{self._classname} disposing [d]")
        if self._cxn is not None:
            self._cxn.remove_listener(
                self._event.refresh_by_id, self._on_stream_refresh
            )
            self._cxn.remove_listener(self._event.update_by_id, self._on_stream_update)
            self._cxn.remove_listener(self._event.status_by_id, self._on_stream_status)
            self._cxn.remove_listener(
                self._event.complete_by_id, self._on_stream_complete
            )
            self._cxn.remove_listener(self._event.error_by_id, self._on_stream_error)
            self._cxn.remove_listener(self._event.ack_by_id, self._on_stream_ack)
            self._release_cxn()
        self._contributed.set()
        self._debug(f"{self._classname} disposed [D]")

    def _do_on_stream_refresh(self, cxn: "StreamConnection", message: dict, *_) -> Any:
        self._refresh_message = message
        message_state = message.get("State", {})
        self._stream_state = message_state.get("Stream", "")
        self._message_state = message_state.get("Text", "")
        return message

    def _on_stream_update(self, originator, *args) -> None:
        self._propagate_event(StreamEvent.UPDATE, originator, *args)

        if self.is_contributing:
            self._update_event.set()

    def _on_stream_status(self, originator, *args) -> None:
        self._propagate_event(StreamEvent.STATUS, originator, *args)

        if self.is_open:
            self.dispatch_complete(originator, *args)

    def _do_on_stream_status(self, cxn: "StreamConnection", message: dict, *_) -> Any:
        self._status_message = message
        message_state = message.get("State", {})
        stream_state = message_state.get("Stream", "")
        self._stream_state = stream_state
        self._message_state = message_state.get("Text", "")

        if stream_state == "Closed":
            self._debug(
                f"{self._classname} received a closing message, "
                f"message_state={message_state}, state={self.state}"
            )

            if self.is_opening:
                self._opened.set()

        return message

    def _on_stream_complete(self, originator, *args) -> None:
        self._propagate_event(StreamEvent.COMPLETE, originator, *args)

        if self.is_opening:
            self._opened.set()

    def _do_on_stream_error(
        self, originator: "StreamConnection", message: dict, *args
    ) -> Any:
        self._error_message = message

        if self.is_contributing:
            if message.get("Type") == "Error":
                debug_message = message.get("Debug", {}).get("Message")
                if debug_message:
                    try:
                        debug_dict = json.loads(debug_message)
                        if debug_dict.get("PostID"):
                            self._contrib_response = ErrorContribResponse(message)
                    except json.decoder.JSONDecodeError:
                        self._error(
                            f"Cannot decode Debug message as JSON: {debug_message}"
                        )

            self._error_event.set()
            return (message, *args)

        else:
            return super()._do_on_stream_error(originator, message, *args)

    def _do_on_stream_ack(
        self, originator: "StreamConnection", message: dict, *args
    ) -> Any:
        self._ack_message = message

        if self.is_contributing:
            self._contrib_response = AckContribResponse(message)
            ack_id = message.get("AckID")
            nack_code = message.get("NakCode")

            if ack_id != self.post_id:
                # Received Ack message with wrong ack_id
                self._error(
                    f"{self._classname} Received Ack message "
                    f"with wrong ack_id={ack_id} != post_id={self.post_id}"
                )
            else:
                if nack_code:
                    reason = message.get("Text")
                    self._error(
                        f"{self._classname} received Nack message, "
                        f"ackID={ack_id}, NackCode={nack_code}, reason={reason}"
                    )

                else:
                    self._debug(
                        f"{self._classname} received Ack message, "
                        f"ackID={ack_id}, state={self.state}"
                    )

            self._ack_event.set()
            return (message, *args)

        else:
            return super()._do_on_stream_ack(originator, message, *args)

    def send_open_message(self):
        self.send(self.open_message)

    def get_contrib_message(
        self, fields: dict, contrib_type: Union[str, "ContribType", None]
    ) -> dict:
        message = {
            "Ack": True,
            "ID": self.id,
            "Message": {
                "Fields": fields,
                "ID": self.id,
                "Type": contrib_type_enum_arg_parser.get_str(
                    contrib_type if contrib_type else "Update"
                ),
                "Domain": self.domain,
            },
            "PostID": self.post_id,
            "Type": "Post",
            "Domain": self.domain,
        }

        post_user_info = self._post_user_info
        if not post_user_info:
            hostname = socket.gethostname()
            ip_addr = socket.gethostbyname(hostname)
            post_user_info = {"Address": ip_addr, "UserID": os.getpid()}

        message["PostUserInfo"] = post_user_info

        return message

    def get_contrib_error_message(self) -> dict:
        return {
            "Type": "Error",
            "Message": "Contribute failed because of disconnection.",
        }

    def contribute(
        self,
        fields: dict,
        contrib_type: Union[str, "ContribType", None] = None,
        post_user_info: Optional[dict] = None,
    ) -> ContribResponse:
        self._is_contributing = True
        self._post_id = self.get_next_post_id()
        self._post_user_info = post_user_info

        self._error_event.clear()
        self._ack_event.clear()
        self._update_event.clear()

        sent = self.send(self.get_contrib_message(fields, contrib_type))

        if sent:
            self._contributed.wait()
            self._is_contributing = False

            if self.is_close:
                self._contrib_response = NullContribResponse()
                self._on_stream_error(self, self.get_contrib_error_message())

        else:
            self._is_contributing = False
            self._contrib_response = NullContribResponse()

        return self._contrib_response

    async def contribute_async(
        self,
        fields: dict,
        contrib_type: Union[str, "ContribType", None] = None,
        post_user_info: Optional[dict] = None,
    ) -> ContribResponse:
        return await asyncio.get_event_loop().run_in_executor(
            None, partial(self.contribute, fields, contrib_type, post_user_info)
        )



from . import bulk
from .._content_data import Data
from ..._content_type import ContentType
from ..._tools import create_repr
from ...delivery._data._data_provider import DataProviderLayer, BaseResponse

package_name_by_content_type = {
    ContentType.ESG_FULL_MEASURES: "esg.full_measures",
    ContentType.ESG_FULL_SCORES: "esg.full_scores",
    ContentType.ESG_STANDARD_MEASURES: "esg.standard_measures",
    ContentType.ESG_STANDARD_SCORES: "esg.standard_scores",
}


def get_package_name(content_type: ContentType):
    package_name = package_name_by_content_type.get(content_type)

    if not package_name:
        raise ValueError(f"Cannot find package_name by content_type={content_type}")

    return package_name


class BaseDefinition(DataProviderLayer[BaseResponse[Data]]):
    def get_db_data(self):
        """
        Returns a response to the data platform

        Returns
        -------
        Response

        """
        definition = bulk.Definition(
            package_name=get_package_name(self._content_type),
            universe=self._kwargs.get("universe"),
        )
        response = definition.get_db_data()
        return response

    def __repr__(self):
        return create_repr(
            self,
            content=f"{{universe='{self._kwargs.get('universe')}', "
            f"start='{self._kwargs.get('start')}', "
            f"end='{self._kwargs.get('end')}', "
            f"closure='{self._kwargs.get('closure')}'}}",
        )



from tests.unit.conftest import args

from tests.unit.content.ipa import data_for_tests as dt

"""
"": args(
    defn=,
    cnt_data=,
    err_msg=,
    err_code=
)
"""
ERROR_MESSAGE_TEST_CASES = {
    "zc_curve_error_message": args(
        defn=dt.ZC_CURVE_DEFINITION,
        cnt_data=dt.ZC_CURVE_RESPONSE,
        err_msg="Validation error",
        err_code=400,
    ),
    "forward_curve_error_message": args(
        defn=dt.FORWARD_CURVE_DEFINITION,
        cnt_data=dt.FORWARD_CURVE_RESPONSE,
        err_msg="The service failed to find the curve constituents",
        err_code="QPS-Curves.10",
    ),
    "zc_curves_error_message": args(
        defn=dt.ZC_CURVES_DEFINITION,
        cnt_data=dt.ZC_CURVES_RESPONSE,
        err_msg="The service failed to find the curve definition",
        err_code="QPS-Curves.7",
    ),
    "surface_cap_error_message": args(
        defn=dt.SURFACE_CAP_DEFINITION,
        cnt_data=dt.SURFACE_CAP_RESPONSE,
        err_msg="The service failed to build the volatility surface",
        err_code="VolSurf.10300",
    ),
    "cap_floor_error_message": args(
        defn=dt.CAP_FLOOR_DEFINITION,
        cnt_data=dt.CAP_FLOOR_RESPONSE,
        err_msg="Market data error : an internal error occured. Failed to cast RateSurfaceInfoResponse.",
        err_code=None,
    ),
    "cds_error_message": args(
        defn=dt.CDS_DEFINITION,
        cnt_data=dt.CDS_RESPONSE,
        err_msg="Technical error occured.",
        err_code="QPS-DPS.1",
    ),
    "options_error_message": args(
        defn=dt.OPTION_DEFINITION,
        cnt_data=dt.OPTION_RESPONSE,
        err_msg="Technical error occured. No Realtime Data retrieved by DAL. Instrument='FCHI560000L1.p' does not exist!",
        err_code="QPS-DPS.1",
    ),
}



import inspect
import json
from functools import lru_cache

SEPARATOR = "   "


def remove_private_and_magick(methods: list) -> list:
    return [i for i in methods if i[0] != "_"]


def extract_instances_from_dict(module, methods_list):
    if "installed_packages" in methods_list:
        methods_list.remove("installed_packages")

    return {i: getattr(module, i) for i in methods_list}


def name_formatter(name, inst, sep=SEPARATOR):
    inst = extract_module_name(inst)
    if " from " in inst:
        inst = inst.split("from")[0]
    elif len(inst) > 200:
        inst = f"{inst[:100]}....."

    name_str = f"{name}{sep}{inst}"

    return name_str


def extract_module_name(inst):
    if (
        (type(inst) is not str)
        and hasattr(inst, "__module__")
        and "class" not in f"{inst}"
    ):
        inst = f"{inst} located in <{inst.__module__}>"
    return f"{inst}"


@lru_cache()
def walker(modd):
    methods = remove_private_and_magick(dir(modd))
    methods = extract_instances_from_dict(modd, methods)

    struct = {"inst": f"{name_formatter('', modd, sep='')}", "resources": methods}
    for i in methods.items():
        if inspect.ismodule(i[1]):
            if "refinitiv.data" in i[1].__spec__.name:
                methods[i[0]] = walker(i[1])
            else:
                methods[i[0]] = f"{name_formatter('', i[1], sep='')} [external]"
        else:
            methods[i[0]] = f"{name_formatter('', i[1], sep='')}"

    return struct


def dict_compare(expected_dict, actual_dict):
    for expected_key, expected_val in expected_dict.items():
        if isinstance(expected_val, dict):
            try:
                dict_compare(expected_val, actual_dict.get(expected_key))
            except AttributeError:
                raise ValueError()
            except ValueError:
                raise Exception(f"{str(expected_val)};{str(expected_dict)};")

    is_equal = expected_dict.keys() == actual_dict.keys()
    assert is_equal, (
        f"These dicts are different:\n"
        f"Expected keys: {expected_dict.keys()}\n"
        f"Actual keys: {actual_dict.keys()}\n"
        f"expected_dict={expected_dict}\n"
        f"actual_dict={actual_dict}"
    )


def test_public_api():
    # given
    import refinitiv.data as rd

    actual_dict = walker(rd)

    # when
    file_path = "./tests/unit/cur_rdp_model.json"
    with open(file_path, "r") as fp:
        expected_dict = json.load(fp)

    # then
    dict_compare(expected_dict, actual_dict)
